/************************** Start of BITIO.C *************************
 *
 * This utility file contains all of the routines needed to impement
 * bit oriented routines under either ANSI or K&R C.  It needs to be
 * linked with every program used in the entire book.
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include "bitio.h"
//#include "errhand.h"

#define PACIFIER_COUNT 2047

#define OUT_BIT(bf,b) {  *(bf->buffer++) = b; bf->count++; }
#define IN_BIT(bf) *(bf->buffer++)

BIT_FILE *OpenOutputBitFile(char *buffer)
{
    BIT_FILE *bit_file;

    bit_file = (BIT_FILE *) calloc( 1, sizeof( BIT_FILE ) );
    if ( bit_file == NULL )
        return( bit_file );

    bit_file->buffer = buffer;
    bit_file->count = 0;

    bit_file->rack = 0;
    bit_file->mask = 0x80;
    bit_file->pacifier_counter = 0;
    return( bit_file );
}

BIT_FILE *OpenInputBitFile(char *buffer, int count )
{
    BIT_FILE *bit_file;

    bit_file = (BIT_FILE *) calloc( 1, sizeof( BIT_FILE ) );
    if ( bit_file == NULL )
	return( bit_file );

    bit_file->buffer=buffer;
    bit_file->start=buffer;
    bit_file->count=count;

    bit_file->rack = 0;
    bit_file->mask = 0x80;
    bit_file->pacifier_counter = 0;
    return( bit_file );
}

int CloseOutputBitFile( bit_file )
BIT_FILE *bit_file;
{
    int count;

    if ( bit_file->mask != 0x80 )
      OUT_BIT(bit_file,bit_file->rack);

    count = bit_file->count;

    free( (char *) bit_file );

    return count;
}

int CloseInputBitFile( bit_file )
BIT_FILE *bit_file;
{
  int bread = bit_file->buffer-bit_file->start;

  free( (char *) bit_file );

  return bread;
}

void OutputBit( bit_file, bit )
BIT_FILE *bit_file;
int bit;
{
    if ( bit )
        bit_file->rack |= bit_file->mask;

    bit_file->mask >>= 1;
    if ( bit_file->mask == 0 ) {
        OUT_BIT(bit_file,bit_file->rack);
	bit_file->rack = 0;
	bit_file->mask = 0x80;
    }
}

void OutputBits( bit_file, code, count )
BIT_FILE *bit_file;
unsigned long code;
int count;
{
    unsigned long mask;

    mask = 1L << ( count - 1 );
    while ( mask != 0) {
        if ( mask & code )
            bit_file->rack |= bit_file->mask;
        bit_file->mask >>= 1;
        if ( bit_file->mask == 0 ) {
          OUT_BIT(bit_file,bit_file->rack);
          bit_file->rack = 0;
          bit_file->mask = 0x80;
        }
        mask >>= 1;
    }
}

int InputBit( bit_file )
BIT_FILE *bit_file;
{
    int value;

    if ( bit_file->mask == 0x80 )
        bit_file->rack = IN_BIT(bit_file);

    value = bit_file->rack & bit_file->mask;
    bit_file->mask >>= 1;
    if ( bit_file->mask == 0 )
	bit_file->mask = 0x80;

    return( value ? 1 : 0 );
}

unsigned long InputBits( bit_file, bit_count )
BIT_FILE *bit_file;
int bit_count;
{
    unsigned long mask;
    unsigned long return_value;

    mask = 1L << ( bit_count - 1 );
    return_value = 0;
    while ( mask != 0) {
	if ( bit_file->mask == 0x80 )
            bit_file->rack = IN_BIT(bit_file);

	if ( bit_file->rack & bit_file->mask )
            return_value |= mask;

        mask >>= 1;
        bit_file->mask >>= 1;
        if ( bit_file->mask == 0 )
            bit_file->mask = 0x80;
    }
    return( return_value );
}

/*************************** End of BITIO.C **************************/

