#include "stdinc.h"
#include <assert.h>
#include "buffer.h"
#include "smalloc.h"

#ifdef BUFFER_GECKO2
#include "fragment.h"
#endif

void BufferDump(struct Buffer *buf)
{
  if (buf->Remain>0) {
    if (buf->Callback!=NULL)
      buf->Callback(buf);

    if (buf->OutFile>=0)
      write(buf->OutFile,buf->BufPtr,buf->Remain);

    if (buf->OutSocket!=INVALID_SOCKET)
      send(buf->OutSocket,buf->BufPtr,buf->Remain,0);
  }
}

void BufferDrainConnection(struct Buffer *buf)
{
  while (BufferGetChar(buf)) ;
}

int SocketReadReady(SOCKET sock, int timeout)
{
  fd_set fds;
  int result;
  struct timeval tv;

  tv.tv_sec = timeout;
  tv.tv_usec = 0;

  FD_ZERO(&fds);
  FD_SET(sock,&fds);
  result = select(sock+1,&fds,NULL,NULL,&tv);

  return (result>0);
}

int BufferGetMore(struct Buffer *buf)
{
  int result;

  if ((buf->Abort) || (buf->HadTimeout))
    return -1;

  if (buf->Sock!=INVALID_SOCKET) {
    if ((buf->Timeout>0) && (!SocketReadReady(buf->Sock,buf->Timeout))) {
      buf->HadTimeout=TRUE;
      result=-1;
    } else
      result = recv(buf->Sock,buf->BufStart,buf->Size,0);

    if ((result==SOCKET_ERROR) || (result==0))
      result = -1;

  } else if (buf->File!=-1) {
    result = read(buf->File,buf->BufStart,buf->Size);

    if (result<=0)
      result = -1;
  } else if (buf->UseNfsHandle) {
#ifdef BUFFER_GECKO2
    struct THttpData data;
    int bytestoget = MIN(buf->Size,4096);

    data.Bytes = buf->BufStart;
    data.Size = buf->Size;

    if ((!buf->Gecko2Eof) && (GetBytes(buf->NfsHandle,buf->TotalBytes,bytestoget,&data,0)) && (data.Count>0)) {
      result = data.Count;
      buf->BufPtr = data.ResultPtr;
      buf->Gecko2Eof|=(result<bytestoget);
    } else
      result = -1;
#endif
  } else
    result = -1;

  if (result>0)
    buf->TotalBytes+=result;

  return result;
}

int BufferFill(struct Buffer *buf)
{
  while (buf->Remain<=0) {
    buf->BufPtr = buf->BufStart;

    if (buf->IsEof==TRUE) {
      buf->CurChar = ' ';
      return FALSE;
    }

    buf->Remain = BufferGetMore(buf);
    if (buf->Remain>0) {
      BufferDump(buf);
    } else {
      buf->IsEof = TRUE;
      buf->CurChar = ' ';
      return FALSE;
    }
  }

  return TRUE;
}


int BufferGetChar(struct Buffer *buf)
{
  if ((buf->Remain<=0) && (!BufferFill(buf)))
    return FALSE;

  return (buf->Remain--,buf->CurChar = *(buf->BufPtr++),TRUE);
};

int BufferGetLine(struct Buffer *buf)
{
  int i;

  // read a line of up to Buffer->Size-1
  for (i=0; i<buf->Size-1; i++) {
    if (!BufferHaveChar(buf))
      return FALSE;

    BufferConsumeChar(buf);

    switch (buf->CurChar) {
      case '\n': *buf->LinePtr='\0';
      	         buf->LinePtr = buf->LineStart;
                 return TRUE;

      case '\r': /* do nothing */ ;
      		 break;

      default: *(buf->LinePtr++) = buf->CurChar;
               break;
    }
  }

  // if we're here, then line length is too long, so terminate the line
  // as if it had a \n.
  *buf->LinePtr='\0';
  buf->LinePtr=buf->LineStart;
  return TRUE;
}

void BufferToBeginning(struct Buffer *buf)
{
  buf->Remain+=(buf->BufPtr-buf->BufStart);
  buf->LinePtr=buf->LineStart;
  buf->BufPtr=buf->BufStart;
}

void BufferReset(struct Buffer *buf)
{
  buf->Abort = FALSE;
  buf->LinePtr = buf->LineStart;
  buf->BufPtr = buf->BufStart;
  buf->Remain = 0;
  buf->OutFile = -1;
  buf->IsEof = FALSE;
  buf->TotalBytes = 0;
  buf->HadTimeout = FALSE;
  buf->Gecko2Eof = FALSE;

  *(buf->LinePtr) = '\0';
  *(buf->BufPtr) = '\0';
}

struct Buffer *BufferCreate()
{
  struct Buffer *buf;

  buf = (struct Buffer*) SafeMalloc(sizeof(struct Buffer));

  assert(buf!=NULL);

  memset(buf,0,sizeof(struct Buffer));

  buf->Size = 16384;
  buf->LineStart = (char*) SafeMalloc(buf->Size);
  buf->BufStart = (char*) SafeMalloc(buf->Size);
  buf->Sock = INVALID_SOCKET;
  buf->OutSocket = INVALID_SOCKET;
  buf->File = -1;
  buf->Timeout = 0;
  buf->Callback = NULL;
  buf->UseNfsHandle=FALSE;

  assert(buf->LineStart!=NULL);
  assert(buf->BufStart!=NULL);

  BufferReset(buf);

  return buf;
}

void BufferFree(struct Buffer *buf)
{
  if (buf->LineStart!=NULL) {
    SafeFree(buf->LineStart);
    buf->LineStart = NULL;
  }

  if (buf->BufStart!=NULL) {
    SafeFree(buf->BufStart);
    buf->BufStart = NULL;
  }

  SafeFree(buf);
}
