#include "stdinc.h"
#include "bitio.h"
#include "lzw12.h"
#include "huff.h"
#include "compress.h"
#include "handle.h"
#include "smalloc.h"

static struct TLZWParams LZWCompressParams, LZWExpandParams;
static struct THuffParams Huff8Bit, Huff12Bit;

int CompressLoaded = FALSE;

void CompressInit(void)
{
  FILE *f;
  int i;

  f = fopen("dict.bin","rb");

  if (f!=NULL) {
    fread(&Huff8Bit,1,sizeof(Huff8Bit),f);
    Huff8Bit.Trees = (struct THuffTree*) malloc(sizeof(struct THuffTree)*Huff8Bit.numtrees);
    fread(Huff8Bit.Trees,1,sizeof(struct THuffTree)*Huff8Bit.numtrees,f);
    for (i=0; i<Huff8Bit.numtrees; i++) {
      Huff8Bit.Trees[i].nodes = malloc(Huff8Bit.NodesSize);
      fread(Huff8Bit.Trees[i].nodes,1,Huff8Bit.NodesSize,f);
      Huff8Bit.Trees[i].codes = malloc(Huff8Bit.CodesSize);
      fread(Huff8Bit.Trees[i].codes,1,Huff8Bit.CodesSize,f);
    }

    fread(&Huff12Bit,1,sizeof(Huff12Bit),f);
    Huff12Bit.Trees = (struct THuffTree*) malloc(sizeof(struct THuffTree)*Huff12Bit.numtrees);
    fread(Huff12Bit.Trees,1,sizeof(struct THuffTree)*Huff12Bit.numtrees,f);
    for (i=0; i<Huff12Bit.numtrees; i++) {
      Huff12Bit.Trees[i].nodes = malloc(Huff12Bit.NodesSize);
      fread(Huff12Bit.Trees[i].nodes,1,Huff12Bit.NodesSize,f);
      Huff12Bit.Trees[i].codes = malloc(Huff12Bit.CodesSize);
      fread(Huff12Bit.Trees[i].codes,1,Huff12Bit.CodesSize,f);
    }

    fread(&LZWCompressParams,1,sizeof(LZWCompressParams),f);
    fread(&LZWExpandParams,1,sizeof(LZWExpandParams),f);
    CompressLoaded=TRUE;

    fclose(f);
  }
}

int UrlToHandleLZW(char *url, char *handle, char **metadataptr)
{
  int count;
  struct TLZWParams cp;

  if (!CompressLoaded)
    return FALSE;

  cp = LZWCompressParams;
  count = LZWCompressString(url,&cp,handle+HANDLE_SKIP);

  SET_URL_METHOD(handle,METHOD_LZW);

  (*metadataptr) = handle+HANDLE_SKIP+count;

  return TRUE;
}

int HandleToUrlLZW(char *handle, int length, char **url, char **metaptr)
{
  struct TLZWParams ep;
  int bread;
  char dbuffer[2048];

  ep = LZWExpandParams;
  LZWExpandString(handle+HANDLE_SKIP,length,&ep,dbuffer,&bread);
  (*metaptr) = handle+HANDLE_SKIP+bread;

  *url = SafeStrDup(dbuffer);

  return TRUE;
}

int UrlToHandleLZHuff(char *url, char *handle, char **metadataptr)
{
  int count,count2;
  struct TLZWParams cp;
  char buffer[2048];

  if (!CompressLoaded)
    return FALSE;

  cp = LZWCompressParams;
  count = LZWCompressString(url,&cp,buffer);
  count2 = HuffmanCompressString(buffer,handle+HANDLE_SKIP,&Huff12Bit);

  SET_URL_METHOD(handle,METHOD_LZHUFF);

  (*metadataptr) = handle+HANDLE_SKIP+count2;

  return TRUE;
}

int HandleToUrlLZHuff(char *handle, int length, char **url, char **metaptr)
{
  struct TLZWParams ep;
  int bread,brhuff;
  char buffer[2048],dbuffer[2048];

  HuffmanExpandString(handle+HANDLE_SKIP,buffer,&Huff12Bit,&brhuff);
  (*metaptr) = handle+HANDLE_SKIP+brhuff;

  ep = LZWExpandParams;
  LZWExpandString(buffer,length,&ep,dbuffer,&bread);

  *url = SafeStrDup(dbuffer);

  return TRUE;
}

int UrlToHandleHuff(char *url, char *handle, char **metadataptr)
{
  int count;

  if (!CompressLoaded)
    return FALSE;

  count = HuffmanCompressString(url,handle+HANDLE_SKIP,&Huff8Bit);

  SET_URL_METHOD(handle,METHOD_HUFF);

  (*metadataptr) = handle+HANDLE_SKIP+count;

  return TRUE;
}

int HandleToUrlHuff(char *handle, int length, char **url, char **metaptr)
{
  int bread;
  char dbuffer[2048];

  HuffmanExpandString(handle+HANDLE_SKIP,dbuffer,&Huff8Bit,&bread);
  (*metaptr) = handle+HANDLE_SKIP+bread;

  *url = SafeStrDup(dbuffer);

  return TRUE;
}

int UrlToHandleCopy(char *url, char *handle, char **metadataptr)
{
  int count;

  count = strlen(url)+1;
  memcpy(handle+HANDLE_SKIP,url,count);

  SET_URL_METHOD(handle,METHOD_COPY);

  (*metadataptr) = handle+HANDLE_SKIP+count;

  return TRUE;
}

int HandleToUrlCopy(char *handle, int length, char **url, char **metaptr)
{
  int bread;

  bread=strlen(handle+HANDLE_SKIP)+1;
  (*url)=(char*) SafeMalloc(bread);
  memcpy(*url,handle+HANDLE_SKIP,bread);
  (*metaptr) = handle+HANDLE_SKIP+bread;

  return TRUE;
}
