#include "stdinc.h"
#include "thread.h"
#include <assert.h>
#include "smalloc.h"
#include "fallback.h"
#include "handle.h"

#ifdef WINDOWS
#include "d:\bc\projects\proxy\hash\Hash.h"
#else
#include "hash.h"
#endif

#define MAX_URL_NODES 4096
#define MAX_HEADER_NODES 4096

struct Hash_Table *pHashUrlTable, *pHashHeaderTable;
MUTEX_TYPE MutexUrlTable, MutexHeaderTable;

pUrlNode *UrlTable;
pHeaderNode *HeaderTable;
int UrlTableCount=0;
int HeaderTableCount=0;

int NumberOfNodes=0, NumberOfNodesCreated=0, NumberOfNodesDeleted=0;

void UrlTableInit()
{
  pHashUrlTable = (Hash_Table*) SafeMalloc(sizeof(Hash_Table));
  UrlTableCount=0;

  Hash_InitTable(pHashUrlTable,0,HASH_STRING_KEYS);
  UrlTable = (pUrlNode*) SafeMalloc(sizeof(pUrlNode)*MAX_URL_NODES);
  CREATE_MUTEX(MutexUrlTable);


  pHashHeaderTable = (Hash_Table*) SafeMalloc(sizeof(Hash_Table));
  HeaderTableCount=0;

  Hash_InitTable(pHashHeaderTable,0,HASH_STRING_KEYS);
  HeaderTable = (pHeaderNode*) SafeMalloc(sizeof(pHeaderNode)*MAX_HEADER_NODES);
  CREATE_MUTEX(MutexHeaderTable);
}

void UrlTableLock(void)
{
  MUTEX_LOCK(MutexUrlTable);
}

void UrlTableUnlock(void)
{
  MUTEX_UNLOCK(MutexUrlTable);
}

pUrlNode UrlNodeAlloc(void)
{
  if (UrlTableCount<MAX_URL_NODES) {
    NumberOfNodes++;
    NumberOfNodesCreated++;
    UrlTable[UrlTableCount] = (pUrlNode) SafeMalloc(sizeof(struct TUrlNode));
    UrlTable[UrlTableCount]->Number = UrlTableCount;
    return UrlTable[UrlTableCount++];
  }

  assert(FALSE);
  return NULL;
}

int UrlTableStoreUrl(char *url)
{
  int isnew=FALSE;
  int retval;
  Hash_Entry *result;

  UrlTableLock();

  result = Hash_CreateEntry(pHashUrlTable,url,&isnew);

  if (isnew) {
    pUrlNode u = UrlNodeAlloc();
    Hash_SetValue(result,u->Number);

    u->Valid = TRUE;
    u->Url = SafeStrDup(url);
//    u->Inode = inode==INODE_DONTCARE ? HandleGetNewInode() : inode;
  }

  retval = (int) Hash_GetValue(result);

  UrlTableUnlock();

  return retval;
}

int UrlTableRetrieveUrl(int ident, char **url)
{
  UrlTableLock();

  *url = NULL;

  if ((ident<0) || (ident>=UrlTableCount) || (!UrlTable[ident]->Valid)) {
    UrlTableUnlock();
    return FALSE;
  }

  *url = SafeStrDup(UrlTable[ident]->Url);

  UrlTableUnlock();

  return TRUE;
}

void HeaderTableLock(void)
{
  MUTEX_LOCK(MutexHeaderTable);
}

void HeaderTableUnlock(void)
{
  MUTEX_UNLOCK(MutexHeaderTable);
}

pHeaderNode HeaderNodeAlloc(void)
{
  if (HeaderTableCount<MAX_HEADER_NODES) {
    HeaderTable[HeaderTableCount] = (pHeaderNode) SafeMalloc(sizeof(struct THeaderNode));
    HeaderTable[HeaderTableCount]->Number = HeaderTableCount;
    return HeaderTable[HeaderTableCount++];
  }

  assert(FALSE);
  return NULL;
}

int HeaderTableStore(char *headers)
{
  int isnew=FALSE;
  int retval;
  Hash_Entry *result;

  HeaderTableLock();

  result = Hash_CreateEntry(pHashHeaderTable,headers,&isnew);

  if (isnew) {
    pHeaderNode hdr = HeaderNodeAlloc();
    Hash_SetValue(result,hdr->Number);

    hdr->Valid = TRUE;
    hdr->Headers = SafeStrDup(headers);
  }

  retval = (int) Hash_GetValue(result);

  HeaderTableUnlock();

  return retval;
}

int HeaderTableRetrieve(int ident, char **headers)
{
  HeaderTableLock();

  *headers = NULL;

  if ((ident<0) || (ident>=HeaderTableCount) || (!HeaderTable[ident]->Valid)) {
    HeaderTableUnlock();
    return FALSE;
  }

  *headers = SafeStrDup(HeaderTable[ident]->Headers);

  HeaderTableUnlock();

  return TRUE;
}

