#include "stdinc.h"
#include <assert.h>
#include "freq.h"

#define NUMDICT 10
char *dict[NUMDICT] = {"http://www.","http://",".net/",".com/",".gov/",".edu/","cgi-bin/","index",".html",".htm"};
int DictTokenFreq[NUMDICT];

int DictCharFreq[256];
unsigned char DictFreqCharacters[256];
unsigned char DictFreqMap[256];

int FreqTable[256];
unsigned char FreqCharacters[256];
unsigned char FreqMap[256];
int FreqTotalStrings;
int FreqMin, FreqMax, FreqAvg, FreqTotalChars;

int LenHist[1024];
int CompHist[1024];

void FreqInit(void)
{
  memset(DictCharFreq,0,sizeof(DictCharFreq));
  memset(DictTokenFreq,0,sizeof(DictTokenFreq));

  memset(CompHist,0,sizeof(CompHist));
  memset(LenHist,0,sizeof(LenHist));
  memset(FreqTable,0,sizeof(FreqTable));
  FreqTotalStrings=0;
  FreqTotalChars=0;
  FreqMin=65535;
  FreqMax=0;
  FreqAvg=0;
}

int DictFind(char *x, char **newpos)
{
  int i;

  (*newpos)=x;

  for (i=0; i<NUMDICT; i++)
    if ((strncmp(x,dict[i],strlen(dict[i]))==0)) {
      *(newpos)=x+strlen(dict[i])-1;
      return i;
    }

  return -1;
}

void FreqAddString(char *thestring)
{
  unsigned char *s = (unsigned char *) thestring;
  int len = strlen(thestring);
  int dtoken;
  char *newpos;

  while (*s) {
      if ((dtoken=DictFind(s,&newpos))>=0) {
      s=newpos;
      DictTokenFreq[dtoken]++;
    } else
      DictCharFreq[*s]++;

    s++;
  }

  s=(unsigned char *) thestring;
  while (*s) {
    FreqTable[*s]++;
    s++;
  }

  FreqTotalStrings++;

  FreqTotalChars+=len;

  FreqMin = MIN(FreqMin,len);
  FreqMax = MAX(FreqMax,len);

  if (len<1024)
    LenHist[len]++;
}

void FreqShowDict(void)
{
  int i;

  fprintf(stdout,"Dictionary....\n");
  for (i=0; i<NUMDICT; i++)
    fprintf(stdout,"  %12s %d\n",dict[i],DictTokenFreq[i]);
}

void FreqDumpHist(void)
{
  FILE *f = fopen("lenhist","wt");
  int i;
  int cumulative=0,ccomp=0,ctotal=0,total=0;

  fprintf(f,"cumulative results\n");

  for (i=0; i<1024; i++) {
    total+=LenHist[i];
    ctotal+=CompHist[i];
  }

  total = MAX(total,1);
  ctotal = MAX(ctotal,1);


  for (i=0; i<1024; i++) {
    cumulative+=LenHist[i];
    ccomp+=CompHist[i];

    fprintf(f,"%d %d %d\n",i,(cumulative*100)/total,(ccomp*100)/ctotal);
  }

  fprintf(f,"normal results\n");

  for (i=0; i<1024; i++) {
    int sum = 0;
    int csum = 0;

    fprintf(f,"%d %d %d\n",i,LenHist[i],CompHist[i]);
  }

  fclose(f);
}

void FreqShowFrequenciesPrime(int *FreqTable)
{
  int lower=0;
  int upper=0;
  int space=0;
  int punct=0;
  int digit=0;
  int total=0;
  int other=0;
  int slash=0;
  int i;

  for (i=0; i<256; i++) {
    if (islower(i))
      lower+=FreqTable[i];
    else if (isupper(i))
      upper+=FreqTable[i];
    else if (i=='/')
      slash+=FreqTable[i];
    else if (isspace(i))
      space+=FreqTable[i];
    else if (ispunct(i))
      punct+=FreqTable[i];
    else if (isdigit(i))
      digit+=FreqTable[i];
    else
      other+=FreqTable[i];
    total+=FreqTable[i];
  }

  fprintf(stdout,"Total URLs.... %d\n",FreqTotalStrings);
  fprintf(stdout,"Total Chars... %d\n",total);
  fprintf(stdout,"String Length. min=%d max=%d avg=%d\n",FreqMin,FreqMax,FreqAvg);
  fprintf(stdout,"Lower Case.... %d (%d%)\n",lower,(lower*100)/total);
  fprintf(stdout,"Upper Case.... %d (%d%)\n",upper,(upper*100)/total);
  fprintf(stdout,"Slash......... %d (%d%)\n",slash,(slash*100)/total);
  fprintf(stdout,"Digits........ %d (%d%)\n",digit,(digit*100)/total);
  fprintf(stdout,"Punctuation... %d (%d%)\n",punct,(punct*100)/total);
  fprintf(stdout,"Space......... %d (%d%)\n",space,(space*100)/total);
  fprintf(stdout,"Other......... %d (%d%)\n",other,(other*100)/total);
}

void FreqQuickSort(int l, int r, int *FreqTable, unsigned char *FreqCharacters)
{
  int i=l,j=r;
  int x;
  unsigned char y;

  x=FreqTable[FreqCharacters[(i+j)/2]];

  do {
    while (FreqTable[FreqCharacters[i]]>x) i++;
    while (x>FreqTable[FreqCharacters[j]]) j--;

    if (i<=j) {
      y=FreqCharacters[i];
      FreqCharacters[i]=FreqCharacters[j];
      FreqCharacters[j]=y;
      i++;
      j--;
    }

  } while (i<=j);

  if (l<j) FreqQuickSort(l,j,FreqTable,FreqCharacters);
  if (i<r) FreqQuickSort(i,r,FreqTable,FreqCharacters);
}


void FreqSort(void)
{
  int x;

  for (x=0; x<256; x++)
    FreqCharacters[x]=(unsigned char) x;

  FreqQuickSort(0,255,FreqTable,FreqCharacters);

  for (x=0; x<256; x++)
    FreqMap[FreqCharacters[x]]=x;

  for (x=0; x<256; x++)
    DictFreqCharacters[x]=(unsigned char) x;

  FreqQuickSort(0,255,DictCharFreq,DictFreqCharacters);

  for (x=0; x<256; x++)
    DictFreqMap[DictFreqCharacters[x]]=x;
}

void FreqCumulativeReport(int *FreqTable, unsigned char *FreqCharacters)
{
  int top31=0;
  int top63=0;
  int total=0;
  int i;

  for (i=0; i<30; i++)
    top31+=FreqTable[FreqCharacters[i]];

  for (i=0; i<62; i++)
    top63+=FreqTable[FreqCharacters[i]];

  for (i=0; i<256; i++)
    total+=FreqTable[i];

  fprintf(stdout,"Top30......... ");
  for (i=0; i<30; i++)
    fprintf(stdout,"%c",FreqCharacters[i]);
  fprintf(stdout,"\n");

  fprintf(stdout,"Cumulat top 30 %d (%d%)\n",top31,(top31*100)/total);

  fprintf(stdout,"Top62......... ");
  for (i=0; i<62; i++)
    fprintf(stdout,"%c",FreqCharacters[i]);
  fprintf(stdout,"\n");

  fprintf(stdout,"Cumulat top 62 %d (%d%)\n",top63,(top63*100)/total);
}

void FreqReadFile(char *fn)
{
  FILE *f = fopen(fn,"rt");
  char buf[256];

  if (f!=NULL) {
    while (fgets(buf,sizeof(buf),f)!=NULL) {
      *(buf+strlen(buf)-1)='\0';
      FreqAddString(buf);
    }
    fclose(f);
  }

  FreqSort();

  FreqAvg = (FreqTotalStrings==0) ? 0 : FreqTotalChars/FreqTotalStrings;
}

void OutCode(char **destptr, int *bitpos, int data, int count)
{
  int i;

  for (i=count-1; i>=0; i--) {
    int thebit = (data & (1<<i))!=0;

    if (thebit)
      (**destptr)|=(1<<*bitpos);

    (*bitpos)--;
    if (*bitpos==-1) {
      *bitpos=7;
      (*destptr)++;
      (**destptr)=0;
    }
  }
}

int GetCode(char **destptr, int *bitpos, int count)
{
  int i;
  int val=0;

  for (i=count-1; i>=0; i--) {
    int thebit= ((**destptr) & (1<<*bitpos))!=0;

    if (thebit)
      val|=(1<<i);

    (*bitpos)--;
    if (*bitpos==-1) {
      *bitpos=7;
      (*destptr)++;
    }
  }

  return val;
}

void Decompress8to5(char *input, int len, char *dest)
{
  char *inptr=input;
  int bitpos=7;
  int code;

  while ((code=GetCode(&inptr,&bitpos,5))!=31) {
    if (code<30)
      *(dest++) = FreqCharacters[code];
    else
      *(dest++) = GetCode(&inptr,&bitpos,8);
  }

  *dest='\0';
}

void Compress8to5(char *thestring, char *dest, int *len)
{
  char *destptr=dest;
  int bitpos=7;
  unsigned char *x = (unsigned char*) thestring;

  *destptr=0;

  while (*x) {
    int pos = FreqMap[*x];

    if (pos<30)
      OutCode(&destptr,&bitpos,pos,5);
    else {
      OutCode(&destptr,&bitpos,30,5);
      OutCode(&destptr,&bitpos,*x,8);
    }
    x++;
  }

  OutCode(&destptr,&bitpos,31,5);

  *len = (bitpos==7) ? (destptr-dest) : (destptr-dest)+1;
}

void Decompress8to6(char *input, int len, char *dest)
{
  char *inptr=input;
  int bitpos=7;
  int code;

  while ((code=GetCode(&inptr,&bitpos,6))!=63) {
    if (code<62)
      *(dest++) = FreqCharacters[code];
    else
      *(dest++) = GetCode(&inptr,&bitpos,8);
  }

  *dest='\0';
}

void Compress8to6(char *thestring, char *dest, int *len)
{
  char *destptr=dest;
  int bitpos=7;
  unsigned char *x = (unsigned char*) thestring;

  *destptr=0;

  while (*x) {
    int pos = FreqMap[*x];

    if (pos<62)
      OutCode(&destptr,&bitpos,pos,6);
    else {
      OutCode(&destptr,&bitpos,62,6);
      OutCode(&destptr,&bitpos,*x,8);
    }
    x++;
  }

  OutCode(&destptr,&bitpos,63,6);

  *len = (bitpos==7) ? (destptr-dest) : (destptr-dest)+1;
}

void Decompress8to6Special(char *input, int len, char *dest)
{
  char *inptr=input;
  int bitpos=7;
  int code;

  while ((code=GetCode(&inptr,&bitpos,6))!=63) {
    if (code==62)
      *(dest++) = GetCode(&inptr,&bitpos,8);
    else if (code>=62-NUMDICT) {
      strcpy(dest,dict[code-(62-NUMDICT)]);
      dest+=strlen(dict[code-(62-NUMDICT)]);
    } else
      *(dest++) = DictFreqCharacters[code];
  }

  *dest='\0';
}

void Compress8to6Special(char *thestring, char *dest, int *len)
{
  char *destptr=dest;
  int bitpos=7;
  unsigned char *x = (unsigned char*) thestring;

  *destptr=0;

  while (*x) {
    int pos = DictFreqMap[*x];
    int found = FALSE;
    int i;

    for (i=0; i<NUMDICT; i++)
      if ((!found) && (strncmp(x,dict[i],strlen(dict[i]))==0)) {
        OutCode(&destptr,&bitpos,62-NUMDICT+i,6);
        found=TRUE;
        x+=strlen(dict[i])-1;
      }

    if (found)
      ; // do nothing
    else if (pos<62-NUMDICT)
      OutCode(&destptr,&bitpos,pos,6);
    else {
      OutCode(&destptr,&bitpos,62,6);
      OutCode(&destptr,&bitpos,*x,8);
    }
    x++;
  }

  OutCode(&destptr,&bitpos,63,6);

  *len = (bitpos==7) ? (destptr-dest) : (destptr-dest)+1;
}

void FreqShowFrequencies(void)
{
  FreqShowFrequenciesPrime(FreqTable);
  fprintf(stdout,"\n--- WITH DICTIONARY ----\n");
  FreqShowFrequenciesPrime(DictCharFreq);
}




