#include "stdinc.h"
#include <assert.h>
#include <string.h>
#include "g2hset.h"

#define STATE_TAG 0
#define STATE_DATA 1

void HsetQuickSort(int l, int r, pHeaderSet hset)
{
  int i=l,j=r;
  char *x;
  char *y;

  x=hset->Tag[(i+j)/2];

  do {
    while (strcmp(hset->Tag[i],x)<0) i++;
    while (strcmp(x,hset->Tag[j])<0) j--;

    if (i<=j) {
      y=hset->Tag[i];
      hset->Tag[i]=hset->Tag[j];
      hset->Tag[j]=y;

      y=hset->Data[i];
      hset->Data[i]=hset->Data[j];
      hset->Data[j]=y;
      
      i++;
      j--;
    }

  } while (i<=j);

  if (l<j) HsetQuickSort(l,j,hset);
  if (i<r) HsetQuickSort(i,r,hset);
}

void HsetDecompose(pHeaderSet dest, char *src)
{
  char *x = src;
  int state = STATE_TAG;
  int done = FALSE;
  char *data, *tag;

  dest->NumHeaders=0;

  if (src==NULL)
    return;

  tag=x;
  data=x;

  while (!done) {
    switch (*x) {
      case ':':
        if (state==STATE_TAG) {
          state=STATE_DATA;
          data = (x+1);
          *x='\0';
        }
        break;

      case '\r':
        *x='\0';
        break;

      case '\0':
        done=TRUE;
        // fallthrough to \n

      case '\n':
        *x='\0';
        if ((state==STATE_DATA) && (dest->NumHeaders<MAX_HEADERS)) {
          while (isspace(*tag)) tag++;
          while (isspace(*data)) data++;

          if (*tag) {
            dest->Tag[dest->NumHeaders] = tag;
            dest->Data[dest->NumHeaders] = data;
            dest->NumHeaders++;
          }
        };
        state=STATE_TAG;
        tag=(x+1);
        break;
    } // switch
    x++;
  } // while
}

void HsetCompose(char *dest, pHeaderSet src)
{
  char *x = dest;
  int i;

  for (i=0; i<src->NumHeaders; i++) {
    x=(char*)stpcpy(x,src->Tag[i]);
    *(x++)=':';
    *(x++)=' ';
    x=(char*)stpcpy(x,src->Data[i]);
    *(x++)='\r';
    *(x++)='\n';
  }

  *x='\0';
}

void HsetMerge(pHeaderSet dest, pHeaderSet first, pHeaderSet second)
{
  int i=0,j=0;

  if (first->NumHeaders>0)
    HsetQuickSort(0,first->NumHeaders-1,first);

  if (second->NumHeaders>0)
    HsetQuickSort(0,second->NumHeaders-1,second);

  dest->NumHeaders = 0;

  while (((i<first->NumHeaders) || (j<second->NumHeaders)) && (dest->NumHeaders<MAX_HEADERS)) {
    int compare = (i>=first->NumHeaders) ? 1 : (j>=second->NumHeaders) ? -1 : stricmp(first->Tag[i],second->Tag[j]);

/*    if (i>=first->NumHeaders)
      compare=1;
    else if (j>=second->NumHeaders)
      compare=-1;
    else
      compare = stricmp(first->Tag[i],second->Tag[j]);*/

    if (compare==0) {
      dest->Tag[dest->NumHeaders] = second->Tag[j];
      dest->Data[dest->NumHeaders] = second->Data[j];
      i++;
      j++;
    } else if (compare<0) {
      dest->Tag[dest->NumHeaders] = first->Tag[i];
      dest->Data[dest->NumHeaders] = first->Data[i];
      i++;
    } else {
      dest->Tag[dest->NumHeaders] = second->Tag[j];
      dest->Data[dest->NumHeaders] = second->Data[j];
      j++;
    }
    dest->NumHeaders++;
  }
}

void HsetUpdate(char **dest, char *orig, char *update)
{
  struct THeaderSet horig;
  struct THeaderSet hupdate;
  struct THeaderSet hdest;
  int olen = (orig==NULL) ? 0 : strlen(orig);
  int ulen = (update==NULL) ? 0 : strlen(update);

  HsetDecompose(&horig,orig);
  HsetDecompose(&hupdate,update);

  HsetMerge(&hdest,&horig,&hupdate);

  (*dest) = (char*) SafeMalloc(olen+ulen+MAX_HEADERS+4);

  (**dest) = '\0';

  HsetCompose(*dest,&hdest);
}












