#include "stdinc.h"
#include <assert.h>
#include "handle.h"
#include "html.h"
#include "buffer.h"
#include "dircache.h"
#include "g2link.h"
#include "inode.h"
#include "url.h"

#define DOTROOT "..."

void CreateDotRoot(pChildNode *childlist, char *url, int parentinode)
{
  fhandle childhandle;
  char linktext[512];
  char *tmp;
  char *metaptr;

  *linktext='\0';
  tmp = url+7; // skip over http://

  while (*tmp!='\0') { // add a ../ for each path component in parent
    if ((*tmp=='/') && (*(tmp+1)!='\0'))
      strcat(linktext,"../");
    tmp++;
  }
  strcat(linktext,"..");

  UrlToHandle(linktext,InodeCreate(parentinode,FALSE),childhandle,FHSIZE,&metaptr);
  HandleSetType(childhandle,FLAG_SYMLINK);
  DirAddChildToList(childlist,DOTROOT,childhandle);
}

int AddLink(pChildNode *childlist, char *name, char *data, int linknumber, char *url, int parentinode)
{
  char linkbuf[512];

  if (!NormalizeUrlFromUrl(url,data,linkbuf,FALSE))
    return FALSE;

  fprintf(stderr,"%d %s\n",linknumber,linkbuf);

  // eliminate any circular URL's
  if (strcmp(linkbuf,url)==0)
    return FALSE;

  // only add links which are longer than http://
  if (strlen(linkbuf)>6) {
    char linkname[256];
    char linktext[512];
    char *metaptr;
    fhandle childhandle;
    pChildNode child;

    strcpy(linktext,DOTROOT);
    strcat(linktext,linkbuf+6);

    if (linknumber>=0)
      sprintf(linkname,"%s%03d",name,linknumber);
    else
      strcpy(linkname,name);

    fprintf(stderr,"%s %s\n",linkname,linktext);

    UrlToHandle(linktext,InodeCreate(parentinode,FALSE),childhandle,FHSIZE,&metaptr);
    HandleSetType(childhandle,FLAG_SYMLINK);

    child = DirAddChildToList(childlist,linkname,childhandle);
    child->Flags|=CHILD_FLAG_ISDOTLINK;
  }

  return TRUE;
}

void ExtractLink(pChildNode *childlist,
                 struct tnode *html,
                 int *nlink,
                 int *nframe,
                 int *nimage,
                 char *url,
                 int parentinode)
{
  struct tparam *p = html->Params;
  char *search, *name;
  int *linknumber;

  if /*((!Preferences.LinkImages) &&*/ (html->Tag==tagIMG) //)
    return;

  switch (html->Tag) {
   case tagFrame:
                search = "src";
                name = ".frame";
                linknumber = nframe;
                break;

   case tagIMG: search = "src";
   		name = ".image";
                linknumber = nimage;
                break;

   case tagAnchor: search = "href";
   		   name = ".link";
                   linknumber = nlink;
                   break;
 }

  /* look for an href= field */
  while ((p!=NULL) && ((p->Ident==NULL) || (stricmp(p->Ident,search)!=0))) {
    p=p->Next;
  }

  /* exit if there is no link */
  if ((p==NULL) || (p->Data==NULL) || (*p->Data=='\0'))
    return;

  if (AddLink(childlist,name,p->Data,*linknumber,url,parentinode))
    (*linknumber)++;
}

void ExtractLinks(pChildNode *childlist, struct tnode *html, int *nlink, int *nframe, int *nimage, char *url, int parentinode)
{
  if (childlist==NULL)
    return;

  while (html!=NULL) {
    if ((html->Tag==tagAnchor) || (html->Tag==tagFrame) || (html->Tag==tagIMG))
      ExtractLink(childlist,html,nlink,nframe,nimage,url,parentinode);

    if (html->Child!=NULL)
      ExtractLinks(childlist,html->Child,nlink,nframe,nimage,url,parentinode);

    html=html->Next;
  }
}