#include "stdinc.h"
#include "thread.h"
#include "handle.h"
#include "fallback.h"
#include "compress.h"

char *UrlMethodNames[] = {"Fallback","LZW","Huffman","LZW-Huffman","Copy"};
int UrlMethodCounts[] = {0,0,0,0,0,0,0,0};

int InodeCounter = 0;

int HandleGetNewInode(void)
{
  InodeCounter++;

  if (InodeCounter<0)
    InodeCounter=1;

  return InodeCounter;
}

void GetReadableHandle(fhandle h, char *dest)
{
  char digits[16] = "0123456789abcdef";
  unsigned char *x = h;
  int i;

  for (i=0; i<FHSIZE; i++) {
    *(dest++) = digits[(*x)>>4];
    *(dest++) = digits[(*x)&0xf];

    x++;
  }

  *dest = '\0';
}

void ShowHandle(FILE *fout, fhandle h)
{
  char tmp[256];

  GetReadableHandle(h,tmp);
  fprintf(fout,"%s\n",tmp);
}

void HandleInitFileType(char *handle, char *name)
{
  if (strncmp(name,".contents",9)==0)
    SET_TYPE(handle,FLAG_CONTENTS);
  else if (strncmp(name,".headers",8)==0)
    SET_TYPE(handle,FLAG_HEADERS);
  else if (strncmp(name,".link",5)==0)
    SET_TYPE(handle,FLAG_SYMLINK);
  else if (strncmp(name,".generator",10)==0)
    SET_TYPE(handle,FLAG_GENERATOR);
  else if (strncmp(name,".info",5)==0)
    SET_TYPE(handle,FLAG_INFO);
  else
    SET_TYPE(handle,FLAG_HARDLINK);
}


int HandleToUrlFallback(char *handle, int length, char **url, char **metaptr)
{
  int *ident = (int*) (handle+5);

  *metaptr = handle+9;

  return UrlTableRetrieveUrl(*ident,url);
}

int HandleToInode(char *handle, int length, int *inode)
{
  int *inodeptr = (int *) (handle+1);

  (*inode) = *inodeptr;

  return TRUE;
}

void InodeToHandle(char *handle, int length, int inode)
{
  int *inodeptr = (int *) (handle+1);

  *inodeptr = inode;
}

int HandleToUrl(char *handle, int length, char **url, int *inode, char **metaptr)
{
  int urlmethod = GET_URL_METHOD(handle);

  *url = NULL;

  HandleToInode(handle,length,inode);

  switch (urlmethod) {
    case METHOD_FALLBACK:
      return HandleToUrlFallback(handle,length,url,metaptr);

    case METHOD_LZW:
      return HandleToUrlLZW(handle,length,url,metaptr);

    case METHOD_LZHUFF:
      return HandleToUrlLZHuff(handle,length,url,metaptr);

    case METHOD_HUFF:
      return HandleToUrlHuff(handle,length,url,metaptr);

    case METHOD_COPY:
      return HandleToUrlCopy(handle,length,url,metaptr);

    default:
      return FALSE;
  }
}

void UrlToHandleFallback(char *url, char *handle, char **metadataptr)
{
  int *ident = (int*) (handle+HANDLE_SKIP);

  SET_URL_METHOD(handle,METHOD_FALLBACK);

  *ident = UrlTableStoreUrl(url);
  *metadataptr = (handle+9);
}

#define TRYCOMP(x,method) memset(buffer,0,length); if ((x(url,buffer,&meta)) && ((count=meta-buffer)<bestcount)) {  memcpy(handle,buffer,count); *metadataptr=handle+count; bestcount=count; bestmethod=method; }

void UrlToHandle(char *url, int inode, char *handle, int length, char **metadataptr)
{
  int *inodeptr = (int*) (handle+1);

  char buffer[2048];
  char *meta;
  int count,bestcount,bestmethod;

  bestmethod=METHOD_FALLBACK;
  bestcount=length-HANDLE_OVERHEAD+1;

  if (strlen(url)<256) {
    TRYCOMP(UrlToHandleLZW,METHOD_LZW)
    TRYCOMP(UrlToHandleLZHuff,METHOD_LZHUFF)
    TRYCOMP(UrlToHandleHuff,METHOD_HUFF)
    TRYCOMP(UrlToHandleCopy,METHOD_COPY)
  }

  if (bestmethod==METHOD_FALLBACK) {
    memset(handle,0,length);
    UrlToHandleFallback(url,handle,metadataptr);
  }

  UrlMethodCounts[bestmethod]++;

  if (inode==INODE_DONTCARE)
    (*inodeptr) = HandleGetNewInode();
  else
    (*inodeptr) = inode;
}

int HandleToHeadersFallback(fhandle handle, int length, char *metaptr, char **dest)
{
  int *ident = (int*) (metaptr);

  return HeaderTableRetrieve(*ident,dest);
}

int HandleToHeaders(fhandle handle, int length, char *metaptr, char **dest)
{
  int headermethod = GET_HSET_METHOD(handle);

  *dest = NULL;

  switch (headermethod) {
    case METHOD_FALLBACK:
      return HandleToHeadersFallback(handle,length,metaptr,dest);

    default:
      return FALSE;
  }
}

void HeadersToHandleFallback(char *headers, char *handle, char *metaptr)
{
  int *ident = (int*) (metaptr);

  SET_HSET_METHOD(handle,METHOD_FALLBACK);

  *ident = HeaderTableStore(headers);
}

void HeadersToHandle(char *headers, char *handle, int length, char *metadataptr)
{
  HeadersToHandleFallback(headers,handle,metadataptr);
}

void HandleCopy(fhandle dest, fhandle src)
{
  memcpy(dest,src,FHSIZE);
}


int HandleGetType(char *handle)

{

  return GET_TYPE(handle);

}


void HandleSetType(char *handle, int htype)

{

  SET_TYPE(handle,htype);

}


void HandleGetInfo(fhandle handle, int length, char *buffer, int *len)

{

  char *url, *metaptr;

  int inode;

  char *headers;

  char *bptr;


  HandleToUrl(handle,length,&url,&inode,&metaptr);

  HandleToHeaders(handle,length,metaptr,&headers);


  sprintf(buffer,"url: %s\nmethod: %s (u=%d c=%d)\n\nheaders:\n%s\n",

    url,

    UrlMethodNames[GET_URL_METHOD(handle)],

    strlen(url),metaptr-handle-HANDLE_SKIP,

    headers);


  SafeFree(url);

  SafeFree(headers);


  *len = strlen(buffer);

}

