#include "stdinc.h"
#include <assert.h>
#include "thread.h"
#include "handle.h"
#include "lru.h"
#include "inode.h"

#define MAX_INODE_GROUPS 65535
#define MAX_INODES_PER_GROUP 65535
#define INODESHIFT 16

struct TInodeGroup {
  unsigned int Counter;
  unsigned int Version;
  pLRUNode LRU;
};

struct TInodeGroup InodeGroups[MAX_INODE_GROUPS+1];
pLRUInfo LRUInode;
MUTEX_TYPE MutexInode;
int InodeInitialized=FALSE;

void InodeInit()
{
  int i;

  LRUInode = LRUInit(sizeof(struct LRUNode));

  for (i=0; i<MAX_INODE_GROUPS; i++) {
    pLRUNode n = LRUAllocNode(LRUInode);

    n->Data=(void*) i;

    LRUPromote(LRUInode,n);

    InodeGroups[i].Counter=8;
    InodeGroups[i].Version=0;
    InodeGroups[i].LRU = n;
  }

  CREATE_MUTEX(MutexInode);

  InodeInitialized=TRUE;
}

unsigned int InodeCreate(unsigned int parentinode, int candonewgroup)
{
  unsigned int group;
  unsigned int result;
  struct TInodeGroup *g;

  assert(InodeInitialized);

  MUTEX_LOCK(MutexInode);

  if ((parentinode==ROOTINODE) && (candonewgroup)) {
    pLRUNode n = LRUPopLeastRecent(LRUInode);

    group = (unsigned int) n->Data;
  } else
    group = (unsigned int) (parentinode>>INODESHIFT);

//  group = MAX(0,group); group is unsigned
  group = MIN(MAX_INODE_GROUPS-1,group);

  g = &(InodeGroups[group]);

  g->Counter++;
  if (g->Counter==0) {
    g->Counter=8;
    g->Version++;
  }

  LRUPromote(LRUInode,g->LRU);

  result = (group<<INODESHIFT) | g->Counter;

  MUTEX_UNLOCK(MutexInode);

  return result;
}

unsigned int InodeGetVersion(unsigned int inode)
{
  unsigned int group;
  unsigned int result;

  MUTEX_LOCK(MutexInode);

  group = inode>>INODESHIFT;

//  group = MAX(0,group); group is unsigned
  group = MIN(group,MAX_INODE_GROUPS-1);

  result = InodeGroups[group].Version;

  MUTEX_UNLOCK(MutexInode);

  return result;
}







