#include "stdinc.h"
#include <assert.h>
#include "fragment.h"
#include "fallback.h"
#include "thread.h"
#include "handle.h"
#include "smalloc.h"
#include "stats.h"
#include "preferences.h"
#include "attr.h"
#include "buffer.h"
#include "html.h"
#include "parse.h"
#include "freq.h"
#include "g2hset.h"
#include "bitio.h"
#include "lzw12.h"
#include "huff.h"
#include "inode.h"

#ifndef WINDOWS
#include <pwd.h>
#include "geckonfs.h"
fhandle RootHandle;
struct passwd NobodyPasswordEntry;
SOCKET SocketForHttp;
#endif

#define LZHUFF_INIT_SIZE 256000

struct sPreferences Preferences;

int DumpHist;
int FreqReport;
int MapugTest;
int HsetMergeTest;
int OutDict;
char StrBody[1024], StrHead[1024], StrAttr[1024], StrReadDir[1024];
char MySquidName[1024];
int MySquidPort;

int randrange(int m)
{
#ifdef WINDOWS
  return random(m);
#else
  double r1 = rand();
  int r = r1*((double) m)/((double) RAND_MAX);

  if (r<0) r=0;
  if (r>(m-1)) r=m-1;

  return r;
#endif
}

void WinsockInit(void)
{
#ifdef WINDOWS
  WSADATA winsock_data;
  WORD version_required = 0x0101; /* Version 1.1 */

  if (WSAStartup (version_required, &winsock_data)!=0) {
    printf("unable to open winsock\n");
    exit(1);
  }
#endif
}

void WinsockDone(void)
{
#ifdef WINDOWS
  WSACleanup();
#endif
}

//int FileId = 15360;

#ifdef WINDOWS
void TestThread(void *arg)
#else
void *TestThread(void*arg)
#endif
{
  pHttpData data = (pHttpData) SafeMalloc(sizeof(struct THttpData));
  int tnum = (int) arg;

  data->Size = 8192;
  data->ResultPtr = NULL;
  data->Bytes = (char*) SafeMalloc(8192+1);
  data->Count = 0;

  while (TRUE) {
    int myfile = 15360+randrange(50);
    int offset = randrange(5000);
    int count = randrange(1000);
    char testbuf[8192], url[256];
    fhandle handle;
    char *metaptr;
    char fn[256];
    FILE *f;
    int bread;

    sprintf(url,"http://astronomysite.com/mapug1/15/msg%d.htm",myfile);

    fprintf(stderr,"%d: asking for %s (%d:%d)\n",tnum,url,offset,count);

    UrlToHandle(url,INODE_DONTCARE,handle,FHSIZE,&metaptr);

    GetBytes(handle,offset,count,data,tnum);

#ifdef WINDOWS
    sprintf(fn,"e:\\website\\mapug1\\15\\msg%d.htm",myfile);
#else
    sprintf(fn,"testdata//msg%d.htm",myfile);
#endif

    f=fopen(fn,"rb");
    if (f!=NULL) {
      fseek(f,offset,SEEK_SET);
      bread=fread(testbuf,1,count,f);
      fclose(f);
    } else fprintf(stderr,"can't open: %s",fn);

    if (bread!=data->Count) {
      fprintf(stderr,"%d: count mismatch %s (%d:%d-%d)\n",tnum,handle,offset,bread,data->Count);
      fprintf(stderr,"%s\n",data->ResultPtr);
      exit(1);
      }
    else if (memcmp(testbuf,data->ResultPtr,bread)!=0) {
      fprintf(stderr,"%d: data mismatch %s (%d:%d)\n",tnum,handle,offset,bread);
      exit(1);
      }
    else
      fprintf(stderr,"%d: success: %s\n",tnum,handle);

//    while (TRUE) ;
  }
}

static struct { char *name, *use ; } Args[] = {
  {"-h",          	"Print usage information"},
  {"-head",		"Output http headers (-head http://foo/)"},
  {"-body",		"Output http body (-body http://foo/)"},
  {"-mapug",		"Mapug thread test"},
  {"-attr",		"Show attributes (-attr http://foo/)"},
  {"-readdir",		"Readdir test (-readdir http://foo/)"},
  {"-freq",		"Character frequency report"},
  {"-dumphist",		"Dump histogram of raw URL lengths"},
  {"-hsetmerge",	"Headerset merge (hset1.txt + hset2.txt --> hset3.txt)"},
  {"-outdict",		"Output compression dictionaries"},
  {"-proxyname",	"Name of proxy server (default -proxyname 198.0.0.100)"},
  {"-proxyport",	"Port of proxy server (default -proxyport 3128)"},
  {NULL, NULL}
  } ;

static void usage( char *name ) {
  int f ;

  printf( "usage: %s [flags]\n", name ) ;
  printf( "The following flags are recognized:\n" ) ;
  for( f = 0 ; Args[f].name != NULL ; f++ )
    printf( "flag: %-5s  meaning: %s\n", Args[f].name, Args[f].use ) ;
  }

static void ProcessCommandLine( int argc, char *argv[] ) {
  int a, f ;

  FreqReport=FALSE;
  MapugTest=FALSE;
  DumpHist=FALSE;
  *StrHead='\0';
  *StrBody='\0';
  *StrAttr='\0';
  *StrReadDir='\0';
  strcpy(MySquidName,"198.0.0.100");
  MySquidPort=3128;

  if (argc<2) {
    usage("geckotest");
    exit(1);
  }

  for( a = 1 ; a < argc ; a++ ) {
    char *cmdname = argv[a];

    for( f = 0 ; Args[f].name != NULL ; f++ )
      if( stricmp(Args[f].name, cmdname) == 0 ) break ;

    switch( f ) {
      case 0: usage("server");
      	      exit(1);
              break; /* usage information */

      case 1: if ((a+1<argc) && (*argv[++a]!='\0'))
                strcpy(StrHead,argv[a]);
	      break;

      case 2: if ((a+1<argc) && (*argv[++a]!='\0'))
                strcpy(StrBody,argv[a]);
	      break;

      case 3: MapugTest=TRUE;
	      break;

      case 4: if ((a+1<argc) && (*argv[++a]!='\0'))
                strcpy(StrAttr,argv[a]);
	      break;

      case 5: if ((a+1<argc) && (*argv[++a]!='\0'))
                strcpy(StrReadDir,argv[a]);
      	      break;

      case 6: FreqReport=TRUE;
      	      break;

      case 7: DumpHist=TRUE;
      	      break;

      case 8: HsetMergeTest=TRUE;
      	      break;

      case 9: OutDict=TRUE;
      	      break;

      case 10: if ((a+1<argc) && (*argv[++a]!='\0'))
                strcpy(MySquidName,argv[a]);
      	      break;

      case 11: if ((a+1<argc) && (*argv[++a]!='\0'))
                MySquidPort=atoi(argv[a]);
      	      break;

    }
  }
}

void HeadTest(void)
{
  pHttpData data = HttpDataAlloc(HTTPDATA_DEFAULT_SIZE);
  char *url = StrHead;
  fhandle handle;
  char *metaptr;

  fprintf(stderr,"%d: asking for headers of %s\n",0,url);

  UrlToHandle(url,INODE_DONTCARE,handle,FHSIZE,&metaptr);
  HeadersToHandle("foo: bar\r\n",handle,FHSIZE,metaptr);

  if (GetHeaders(handle,0,4096,data,0)) {
    *(data->ResultPtr+data->Count)='\0';
    fprintf(stdout,"%s",data->ResultPtr);
  } else
    fprintf(stdout,"failure!\n");

  HttpDataFree(data);
}

void BodyTest(void)
{
  pHttpData data = HttpDataAlloc(HTTPDATA_DEFAULT_SIZE);
  char *url = StrBody;
  fhandle handle;
  char *metaptr;
  int bread;
  int offset=0;
  int i;

  memset(data->Bytes,0,data->Size+1);

  UrlToHandle(url,INODE_DONTCARE,handle,FHSIZE,&metaptr);
  HeadersToHandle("foo: bar\r\n",handle,FHSIZE,metaptr);

  do {
    bread = GetBytes(handle,offset,data->Size,data,0) ? data->Count : 0;
    fprintf(stderr,"%d\n",bread);

    if (bread>0) {
      *(data->ResultPtr+bread)='\0';
      fprintf(stdout,"%s",data->ResultPtr);
    }

    offset+=bread;
  } while (bread==data->Size);

  HttpDataFree(data);
}

void AttrTest(void)
{
  pHttpData data = HttpDataAlloc(HTTPDATA_DEFAULT_SIZE);
  char *url = StrAttr;
  fhandle handle;
  char *metaptr;
  struct fattr attr;

  memset(data->Bytes,0,data->Size+1);

  fprintf(stderr,"%d: asking for attributes of %s\n",0,url);

  UrlToHandle(url,INODE_DONTCARE,handle,FHSIZE,&metaptr);
  HeadersToHandle("foo: bar\r\n",handle,FHSIZE,metaptr);

  SET_TYPE(handle,FLAG_CONTENTS);

  if (AttrLookupConsistant(handle,&attr))
    fprintf(stdout,"Type: %d\nMode: %o\nnlink: %d\nuid: %d\ngid: %d\nsize: %d\nblocksize: %d\nrdev: %d\nblocks: %d\nfsid: %d\nfileid: %d\n",
      attr.type,attr.mode,attr.nlink,attr.uid,attr.gid,attr.size,attr.blocksize,attr.rdev,attr.blocks,attr.fsid,attr.fileid);
  else
    fprintf(stdout,"failure!\n");

  HttpDataFree(data);
}

void ReadDirTest(void)
{
  pHttpData data = HttpDataAlloc(HTTPDATA_DEFAULT_SIZE);
  char *url = StrReadDir;
  fhandle handle;
  char *metaptr;
  pentry first=NULL;

  memset(data->Bytes,0,data->Size+1);

  UrlToHandle(url,INODE_DONTCARE,handle,FHSIZE,&metaptr);
  SET_TYPE(handle,FLAG_HARDLINK);

  HeadersToHandle("foo: bar\r\n",handle,FHSIZE,metaptr);

  ShowHandle(stderr,handle);

  if (!DirMakeConsist(handle))
    fprintf(stderr,"consist failed\n");

  if (DirRead(handle,0,&first)<0)
    fprintf(stderr,"dirread failed\n");

  while (first!=NULL) {
    fprintf(stdout,"%d: %s inode=%d\n",*((int*)&first->cookie),first->name,first->fileid);
    first=first->nextentry;
  }
}

struct THuffParams Huff8Bit, Huff12Bit;

void HuffInit()
{
  char compressinit[131072+1];
  int bread=0;
  FILE *f;

  static int HuffAlready = FALSE;

  if (HuffAlready)
    return;

  HuffAlready=TRUE;

  f=fopen("lzwinit.log","rb");
  if (f!=NULL) {
    bread=fread(compressinit,1,131072,f);
    fclose(f);
  }
  *(compressinit+bread)='\0';

  HuffmanInit(&Huff8Bit,256,0,8,257);
  HuffmanConfigure(&Huff8Bit,compressinit);
}


struct TLZWParams LZWCompressParams, LZWExpandParams;

void LZWInit()
{
  int count;
  FILE *f;
  char dest[32768+1];
  char test2[32768+1];
  char compressinit[32768+1];
  int bread=0,junk=0;

  static int LZWAlready = FALSE;

  if (LZWAlready)
    return;

  LZWAlready=TRUE;

  f=fopen("lzwinit.log","rb");
  if (f!=NULL) {
    bread=fread(compressinit,1,32768,f);
    fclose(f);
  }
  *(compressinit+bread)='\0';

  LZWCompressInit(&LZWCompressParams);
  count=LZWCompressString(compressinit,&LZWCompressParams,dest);

  LZWExpandInit(&LZWExpandParams);
  LZWExpandString(dest,count,&LZWExpandParams,test2,&junk);
}

void LZHuffInit()
{
  char compressinit[LZHUFF_INIT_SIZE+1];
  char dest[LZHUFF_INIT_SIZE+1];
  int bread=0;
  FILE *f;
  struct TLZWParams cp,ep;
  int junk;

  char *astring = "http://www.cs.arizona.edu/index.html";
  int c1,c2,c3;
  char dest1[1024],dest2[1024],dest3[1024],dest4[1024];

  static int LZHuffAlready = FALSE;

  if (LZHuffAlready)
    return;

  LZHuffAlready=TRUE;

  f=fopen("lzwinit.log","rb");
  if (f!=NULL) {
    bread=fread(compressinit,1,LZHUFF_INIT_SIZE,f);
    fclose(f);
  }
  *(compressinit+bread)='\0';

  cp=LZWCompressParams;
  LZWCompressString(compressinit,&cp,dest);

  HuffmanInit(&Huff12Bit,8192,256,13,1);
  HuffmanConfigure(&Huff12Bit,dest);

  cp=LZWCompressParams;
  c1=LZWCompressString(astring,&cp,dest1);
  c2=HuffmanCompressString(dest1,dest2,&Huff12Bit);
  c3=HuffmanExpandString(dest2,dest3,&Huff12Bit,&junk);
  ep=LZWExpandParams;
  LZWExpandString(dest3,0,&ep,dest4,&junk);
}

void CompressTest()
{
  char dest[1024],dest2[1024],dest3[1024];
  char inbuf[1024];
  char test2[1024];
  int len8to5, len8to6, len8to6s, lenlzw, lenhuff, lenmin, lenlzhuff;
  int t8to5=0, t8to6=0, t8to6s=0, tlzw=0, thuff=0, tmin=0, tlzhuff=0;
  int c8to5=0, c8to6=0, c8to6s=0, clzw=0, chuff=0, clzhuff=0;
  int totalurl=0;
  int totalbytes=0;
  FILE *f;
  struct TLZWParams cp, ep;
  int junk;

  LZWInit();
  HuffInit();
  LZHuffInit();

  f = fopen("urlmin.log","rt");
  if (f!=NULL) {
    while (fgets(inbuf,sizeof(inbuf),f)!=NULL) {
      *(inbuf+strlen(inbuf)-1)='\0';

      totalbytes+=strlen(inbuf)+1; //+1 for null

      Compress8to5(inbuf,dest,&len8to5);
      Decompress8to5(dest,len8to5,test2);

      if (strcmp(inbuf,test2)!=0)
        assert(FALSE);

      Compress8to6(inbuf,dest,&len8to6);
      Decompress8to6(dest,len8to6,test2);

      if (strcmp(inbuf,test2)!=0)
        assert(FALSE);

      Compress8to6Special(inbuf,dest,&len8to6s);
      Decompress8to6Special(dest,len8to6s,test2);

      if (strcmp(inbuf,test2)!=0)
        assert(FALSE);

      cp=LZWCompressParams;
      lenlzw = LZWCompressString(inbuf,&cp,dest);
      ep=LZWExpandParams;
      LZWExpandString(dest,lenlzw,&ep,test2,&junk);

      if (strcmp(inbuf,test2)!=0)
        assert(FALSE);

      lenlzhuff=HuffmanCompressString(dest,dest2,&Huff12Bit);
      HuffmanExpandString(dest2,dest3,&Huff12Bit,&junk);
      ep=LZWExpandParams;
      LZWExpandString(dest3,-1,&ep,test2,&junk);

      if (strcmp(inbuf,test2)!=0)
        assert(FALSE);

      lenhuff=HuffmanCompressString(inbuf,dest,&Huff8Bit);
      HuffmanExpandString(dest,test2,&Huff8Bit,&junk);

      if (strcmp(inbuf,test2)!=0)
        assert(FALSE);

      totalurl++;
      lenmin=MIN(len8to5,MIN(len8to6,MIN(len8to6s,MIN(lenlzw,MIN(lenhuff,lenlzhuff)))));

      if (lenmin==lenlzhuff)
        clzhuff++;
      else if (lenmin==lenlzw)
        clzw++;
      else if (lenmin==lenhuff)
        chuff++;
      else if (lenmin==len8to6s)
        c8to6s++;
      else if (lenmin==len8to6)
        c8to6++;
      else if (lenmin==len8to5)
        c8to5++;

      t8to5+=len8to5;
      t8to6+=len8to6;
      t8to6s+=len8to6s;
      tlzw+=lenlzw;
      thuff+=lenhuff;
      tmin+=lenmin;
      tlzhuff+=lenlzhuff;

      CompHist[lenlzhuff]++;
    }
    fclose(f);
  }

  totalbytes=MAX(totalbytes,1);

  fprintf(stdout,"Total bytes... %6d\n",totalbytes);
  fprintf(stdout,"8 to 5........ %6d (%02d%%) %4d (%02d%%)\n",t8to5,t8to5*100/totalbytes,c8to5,c8to5*100/totalurl);
  fprintf(stdout,"8 to 6........ %6d (%02d%%) %4d (%02d%%)\n",t8to5,t8to6*100/totalbytes,c8to6,c8to6*100/totalurl);
  fprintf(stdout,"8 to 6 special %6d (%02d%%) %4d (%02d%%)\n",t8to6s,t8to6s*100/totalbytes,c8to6s,c8to6s*100/totalurl);
  fprintf(stdout,"lzw........... %6d (%02d%%) %4d (%02d%%)\n",tlzw,tlzw*100/totalbytes,clzw,clzw*100/totalurl);
  fprintf(stdout,"huffman....... %6d (%02d%%) %4d (%02d%%)\n",thuff,thuff*100/totalbytes,chuff,chuff*100/totalurl);
  fprintf(stdout,"lzwhuff....... %6d (%02d%%) %4d (%02d%%)\n",tlzhuff,tlzhuff*100/totalbytes,clzhuff,clzhuff*100/totalurl);
  fprintf(stdout,"best.......... %6d (%02d%)\n",tmin,tmin*100/totalbytes);
}

void OutputDictionaries(void)
{
  FILE *f;
  int na;
  int i;

  HuffInit();
  LZWInit();
  LZHuffInit();

  f = fopen("dict.bin","wb");

  fwrite(&Huff8Bit,1,sizeof(Huff8Bit),f);
  fwrite(Huff8Bit.Trees,1,sizeof(struct THuffTree)*Huff8Bit.numtrees,f);
  for (i=0; i<Huff8Bit.numtrees; i++) {
    fwrite(Huff8Bit.Trees[i].nodes,1,Huff8Bit.NodesSize,f);
    fwrite(Huff8Bit.Trees[i].codes,1,Huff8Bit.CodesSize,f);
  }

  fwrite(&Huff12Bit,1,sizeof(Huff12Bit),f);
  fwrite(Huff12Bit.Trees,1,sizeof(struct THuffTree)*Huff12Bit.numtrees,f);
  for (i=0; i<Huff12Bit.numtrees; i++) {
    fwrite(Huff12Bit.Trees[i].nodes,1,Huff12Bit.NodesSize,f);
    fwrite(Huff12Bit.Trees[i].codes,1,Huff12Bit.CodesSize,f);
  }

  fwrite(&LZWCompressParams,1,sizeof(LZWCompressParams),f);
  fwrite(&LZWExpandParams,1,sizeof(LZWExpandParams),f);

  fclose(f);
}

void DoHsetMergeTest(void)
{
  char orig[4096];
  char update[4096];
  char *result;
  FILE *f;

  *orig='\0';
  *update='\0';

  f=fopen("hset1.txt","rb");
  if (f!=NULL) {
    fread(orig,1,sizeof(orig),f);
    fclose(f);
  }

  f=fopen("hset2.txt","rb");
  if (f!=NULL) {
    fread(update,1,sizeof(update),f);
    fclose(f);
  }

  HsetUpdate(&result,orig,update);

  f=fopen("hset3.txt","wb");
  if (f!=NULL) {
    fwrite(result,1,strlen(result),f);
    fclose(f);
  }
}

void FreqLoad(void)
{
  static int loaded=FALSE;

  if (!loaded) {
    loaded=TRUE;
    FreqReadFile("urlmin.log");
  }
}

int main(int argc, char **argv)
{
  THREAD_TYPE ThreadID[5];
  int i;

  ProcessCommandLine(argc,argv);

//  randomize();

  InitStatistics();
  CompressInit();
  InodeInit();
  UrlTableInit();
  DirInit();
  FragmentInit(MySquidName,MySquidPort);
  AttrInit();
  WinsockInit();
  ParseInitModule();
  FreqInit();

  if (FreqReport) {
    FreqLoad();
    FreqShowFrequenciesPrime(FreqTable);
    FreqCumulativeReport(FreqTable,FreqCharacters);
    fprintf(stdout,"\n---- With Dictionary ----\n");
    FreqShowFrequenciesPrime(DictCharFreq);
    FreqCumulativeReport(DictCharFreq,DictFreqCharacters);
    FreqShowDict();

    fprintf(stdout,"\n---- Compression Test ----\n");
    CompressTest();
  }

  if (DumpHist) {
    FreqLoad();
    FreqDumpHist();
  }

  if (OutDict)
    OutputDictionaries();

  if (*StrHead)
    HeadTest();

  if (*StrAttr)
    AttrTest();

  if (*StrBody)
    BodyTest();

  if (*StrReadDir)
    ReadDirTest();

  if (HsetMergeTest)
    DoHsetMergeTest();

  if (MapugTest)
    for (i=0; i<8; i++)
      CREATE_THREAD(ThreadID[i],TestThread,(void*) i);

 // while (TRUE) ;

  WinsockDone();

  return 0;
}
