#include "stdinc.h"
#include <assert.h>
#include "smalloc.h"
#include "buffer.h"
#include "html.h"
#include "parse.h"
#ifndef WINDOWS
#ifndef NOTHREAD
#endif
#endif

#define DBUFF_SIZE 262144

#include "thread.h"

#define CurByte (buf->CurChar)
#define allocdbuff (buf->Html_allocdbuff)
#define dbuff (buf->Html_dbuff)
#define dstart (buf->Html_dstart)
#define dend (buf->Html_dend)
#define CurToken (buf->Html_CurToken)
#define LastToken (buf->Html_LastToken)
#define CurTok_Negate (buf->Html_CurTok_Negate)
#define Preform (buf->Html_Preform)

struct LookupMappingNode {
  char *Name;
  int Index;
};

int ParseWantNewlineLeadSpace = FALSE;

struct LookupMappingNode LookupMapping[TokCount];
int LookupCharMapping[256];

int IsLinkTag[TokCount];
int IsSTag[TokCount];
int IsLSTag[TokCount];

char *Lookup[TokCount] = {
         "\0",
          "\0",
          "\0",
         "HTML",
         "TITLE",
         "BODY",
         "H1","H2","H3","H4","H5","H6","H7","H8",
         "P",
         "UL","OL","DL",
         "LI",
         "DD","DT",
         "ADDRESS","B","CITE","DFN","DIR","EM","I","KBD","LISTING",
           "PLAINTEXT","S","SAMP","STRIKE","STRONG","TT","U","VAR",
           "XMP","CODE",
         "CENTER","NOBR","PRE","WBR","BLOCKQUOTE",
         "BR",
         "HR",
         "IMG",
         "FORM","INPUT","ISINDEX",
         "SELECT","OPTION",
         "TABLE","TR","TD","TH",
         "MENU",
         "A",
         "!",
         "META",
         "FRAME",
         "FRAMESET",
         "BASE",
         "MAP",
         "AREA"};

int ModuleInit = FALSE;
MUTEX_TYPE MutexParamFreeList;
MUTEX_TYPE MutexHtmlFreeList;
struct tparam *ParamFreeList = NULL;
struct tnode *HtmlFreeList = NULL;

struct tparam *AllocateParamNode(void)
{
  struct tparam *result;

  MUTEX_LOCK(MutexParamFreeList);

  if (ParamFreeList!=NULL) {
    result = ParamFreeList;
    ParamFreeList = ParamFreeList->Next;
  } else
    result = (struct tparam *) SafeMalloc(sizeof(struct tparam));

  MUTEX_UNLOCK(MutexParamFreeList);

  assert(result!=NULL);
  return result;
}

void FreeParamNode(struct tparam *p)
{
  MUTEX_LOCK(MutexParamFreeList);

  p->Next = ParamFreeList;
  ParamFreeList = p;

  MUTEX_UNLOCK(MutexParamFreeList);
}

struct tnode *AllocateHtmlNode(void)
{
  struct tnode *result;

  MUTEX_LOCK(MutexHtmlFreeList);

  if (HtmlFreeList!=NULL) {
    result = HtmlFreeList;
    HtmlFreeList = HtmlFreeList->Next;
  } else
    result = (struct tnode *) SafeMalloc(sizeof(struct tnode));

  MUTEX_UNLOCK(MutexHtmlFreeList);

  assert(result!=NULL);
  return result;
}

void FreeHtmlNode(struct tnode *p)
{
  MUTEX_LOCK(MutexHtmlFreeList);

  p->Next = HtmlFreeList;
  HtmlFreeList = p;

  MUTEX_UNLOCK(MutexHtmlFreeList);
}

int is_linktag(int tag)
{
  switch (tag) {
    case tagAnchor:
    case tagIMG:
    case tagComment:
    case tagFrame:
    case tagBase:
    case tagMap:
    case tagArea:
    case tagBody:
      return TRUE;

    default:
      return FALSE;
  }
}

int is_stag(int tag)
{
  switch (tag) {
    case tagText:
    case tagPara:
    case tagLI:
    case tagDD:
    case tagDT:
    case tagBR:
    case tagIMG:
    case tagINPUT:
    case tagISINDEX:
    case tagOption:
    case tagHR:
    case tagComment:
    case tagSpace:
    case tagMeta:
    case tagFrame:
    case tagBase:
    case tagArea:
      return TRUE;

    default:
      return FALSE;
  }
}

int is_lstag(int tag)
{
  switch (tag) {
    case tagB:
    case tagAddress:
    case tagCite:
    case tagDFN:
    case tagEM:
    case tagI:
    case tagKBD:
    case tagS:
    case tagStrike:
    case tagStrong:
    case tagTT:
    case tagU:
    case tagVar:
    case tagAnchor:
    case tagCode:
    case tagSamp:
      return TRUE;
    default:
      return FALSE;
  }
}

char *tagname(int tag)
{
  static char buf[256];

  switch (tag) {
    case tagUnknown: sprintf(buf,"unknown"); break;
    case tagText: sprintf(buf,"text"); break;
    case tagSpace: sprintf(buf,"space"); break;
    default: sprintf(buf,"<%s>",Lookup[tag]);
  }

  return buf;
}

void ShowTree(int indent, struct tnode *t)
{
  int i;

  while (t!=NULL) {
  /*
    struct tparam *p = t->Params;

    while (p!=NULL) {
      if ((stricmp(p->Ident,"href")==0) && (*p->Data!='#'))
        printf("%s\n",p->Data);
      p=p->Next;
    } */

    for (i = 0; i<indent; i++) printf(" ");
    printf("%s",tagname(t->Tag));

    if (t->Text!=NULL)
      printf(": %s",t->Text);

    printf("\n");

    if (t->Child!=NULL)
      ShowTree(indent+1,t->Child);

    t=t->Next;
  }
}

static struct { char *name; unsigned char map; } AmpTable[] = {
  {"nbsp",      ' '},
  {"lt",        '<'},
  {"gt",        '>'},
  {"amp",       '&'},
  {"quot",      '"'},
  {"iexcl",     161},
  {"cent",      162},
  {"pound",     163},
  {"curren",    164},
  {"yen",       165},
  {"brvbar",    166},
  {"sect",      167},
  {"uml",       168},
  {"copy",      169},
  {"ordf",      170},
  {"laquo",     171},
  {"not",       172},
  {"shy",       173},
  {"reg",       174},
  {"macr",      175},
  {"deg",       176},
  {"plusmn",    177},
  {"sup2",      178},
  {"sup3",      179},
  {"acute",     180},
  {"micro",     181},
  {"para",      182},
  {"middot",    183},
  {"cedil",     184},
  {"sup1",      185},
  {"ordm",      186},
  {"raquo",     187},
  {"frac14",    188},
  {"frac12",    189},
  {"frac34",    190},
  {"iquest",    191},
  {"Agrave",    192},
  {"Aacute",    193},
  {"Acirc",     194},
  {"Atilde",    195},
  {"Auml",      196},
  {"Aring",     197},
  {"Aelig",     198},
  {"Ccedil",    199},
  {"Egrave",    200},
  {"Eacute",    201},
  {"Ecirc",     202},
  {"Euml",      203},
  {"Igrave",    204},
  {"Iacute",    205},
  {"Icirc",     206},
  {"Iuml",      207},
  {"ETH",       208},
  {"Ntilde",    209},
  {"Ograve",    210},
  {"Oacute",    211},
  {"Ocirc",     212},
  {"Otilde",    213},
  {"Ouml",      214},
  {"times",     215},
  {"Oslash",    216},
  {"Ugrave",    217},
  {"Uacute",    218},
  {"Ucirc",     219},
  {"Uuml",      220},
  {"Yacute",    221},
  {"THORN",     222},
  {"szlig",     223},
  {"agrave",    224},
  {"aacute",    225},
  {"acirc",     226},
  {"atilde",    227},
  {"auml",      228},
  {"aring",     229},
  {"aelig",     230},
  {"ccedil",    231},
  {"egrave",    232},
  {"eacute",    233},
  {"ecirc",     234},
  {"euml",      235},
  {"igrave",    236},
  {"iacute",    237},
  {"icirc",     238},
  {"iuml",      239},
  {"eth",       240},
  {"ntilde",    241},
  {"ograve",    242},
  {"oacute",    243},
  {"ocirc",     244},
  {"otilde",    245},
  {"ouml",      246},
  {"divide",    247},
  {"oslash",    248},
  {"ugrave",    249},
  {"uacute",    250},
  {"ucirc",     251},
  {"uuml",      252},
  {"yacute",    253},
  {"thorn",     254},
  {"yuml",      255},
  {NULL, 0}};


char LookupAmp(char *buf)
{
  int i;

  if (*buf=='#')
    return (char) atoi(buf+1);

  for (i=0; AmpTable[i].name!=NULL; i++)
    if (strcmp(AmpTable[i].name,buf)==0)
      return (char) AmpTable[i].map;

  for (i=0; AmpTable[i].name!=NULL; i++)
    if (stricmp(AmpTable[i].name,buf)==0)
      return (char) AmpTable[i].map;

  return '&';
}

void ExpandAmp(struct Buffer *buf)
{
  char amp[34];
  int count = 0;

  *amp = '\0';
  BufferGetChar(buf);

  while ((CurByte!=';') && (!buf->IsEof) && (count<32) && (!isspace(CurByte)) && (CurByte!='<') && (CurByte!='&')) {
    amp[count++] = (char) toupper(CurByte);
    BufferGetChar(buf);
  }

  if (CurByte!=';') {
    *dend = '&'; /* put back the & */
    return; /* return without consuming another character */
  }

  amp[count]='\0';
  *dend = LookupAmp(amp); /* ought to put the mapped character here! */
  BufferGetChar(buf);
}

void SkipCR(struct Buffer *buf)
{
  while ((!buf->IsEof) && (CurByte=='\r'))
    BufferGetChar(buf);
}

void SkipWhite(struct Buffer *buf)
{
    while ((!buf->IsEof) && ((CurByte=='\r') || (CurByte==' ') || (CurByte=='\n')))
      BufferGetChar(buf);
}

int GetToken(struct Buffer *buf)
{
  int leadspace;
  char *tstart, *tokptr;
  char tok[256];
  int toklen;

  CurTok_Negate = FALSE;

  dstart = dbuff; /* start of text buffer */
  dend = dbuff; /* end of text buffer */

  if (Preform) {
    do {
      SkipCR(buf);
      while ((!buf->IsEof) && (CurByte!='\r') && (CurByte!='\n') && (CurByte!='<')) {
        if (CurByte=='&')
          ExpandAmp(buf);
        else {
          *dend = CurByte;
          BufferGetChar(buf);
        }
        dend++;
      }
    } while ((!buf->IsEof) && (CurByte!='<') && (CurByte!='\n'));

    if ((dstart==dend) && (CurByte == '\n')) {
      BufferGetChar(buf);
      CurToken = tagBR;
      return TRUE;
    }
  } else {
    leadspace = FALSE;

    while ((!buf->IsEof) && ((CurByte == ' ') || (CurByte == '\r') || (CurByte == '\n'))) {
      switch (CurByte) {
        case ' ': leadspace = TRUE;
        	  break;

        case '\n':
        case '\r': leadspace = ParseWantNewlineLeadSpace;
	   	   break;
      }
      BufferGetChar(buf);
    }

    if (leadspace) {
      CurToken = tagSpace;
      return TRUE;
    }

    do {
      SkipWhite(buf);
      while ((!buf->IsEof) && (CurByte!='\r') && (CurByte!='\n') && (CurByte!=' ') && (CurByte!='<')) {
        if (CurByte=='&')
          ExpandAmp(buf);
        else {
          *dend=CurByte;
          BufferGetChar(buf);
        }
        dend++;

        //**** 1/6/00: If the buffer is near full, then return a text token
        //  to avoid causing a buffer overrun and segfault
        if (dend-dstart>DBUFF_SIZE-1024) {
          *dend='\0';
          CurToken=tagText;
          return TRUE;
        }
      }

      /* add up to one space to the end of dend; Consume all remaining spaces */

      if ((!buf->IsEof) && ((CurByte==' ') || (CurByte=='\r') || (CurByte=='\n'))) {
        BufferGetChar(buf);
        *dend=' ';
        dend++;
        while ((!buf->IsEof) && ((CurByte==' ') || (CurByte=='\r') || (CurByte=='\n')))
          BufferGetChar(buf);
      }
    } while ((!buf->IsEof) && (CurByte!='<'));
  }

  if (dend!=dstart) {
    *dend='\0';
    CurToken = tagText;
    return TRUE;
  } else if (CurByte=='<') {
    BufferGetChar(buf);
    while ((!buf->IsEof) && (CurByte!='>')) {
      if ((CurByte=='\r') || (CurByte=='\n'))
        *dend = ' ';
      else
        *dend = CurByte;

      dend++;
      BufferGetChar(buf);
    }
    BufferGetChar(buf);
    *dend = '\0';

    if (*dstart=='/') {
      tstart=dstart+1;
      CurTok_Negate = TRUE;
    } else {
      tstart=dstart;
      CurTok_Negate = FALSE;
    }

    tokptr = tok;
    toklen = 0;
    while ((*tstart!=' ') && (*tstart!='\0') && (toklen<256)) {
      *(tokptr++) = (char) tolower(*tstart);
      tstart++;
      toklen++;
    }

    /* error - toklen>256, something must be horribly wrong! */
    if (toklen>=256)
      return FALSE;

    *tokptr = '\0';

    while (*tstart==' ')
      tstart++;

    dstart = tstart;

    if (*tok=='!')
      CurToken = tagComment;
    else {
      int toknum = LookupCharMapping[(unsigned char) *tok];

      if (toknum!=-1) {
        int match=1;

        while ((toknum<TokCount) && ((match=strcmp(tok,LookupMapping[toknum].Name))>0))
          toknum++;

        if (match!=0)
          toknum=-1;
      }

      CurToken = (toknum!=-1) ? LookupMapping[toknum].Index : tagUnknown;
    }

    return TRUE;
  } else
    return FALSE;
}

int PullData(char **s1, char **dest, char *last)
{
  char *e, *s = *s1;
  int len;
 // char *l;

  /* skip over leading spaces or equal sign */
  while ((*s) && ((isspace(*s)) || (*s=='=')))
    s++;

  e=s;

  /* if it's a quoted string, then get the whole thing */
  if (*s=='"') {
    s++; /* skip over the start quote */
    e++;

    /* skip up to the end quote */
    while ((*e) && (*e!='"'))
      e++;

    /* skip over the end quote */
    if (*e=='"')
      e++;

  } else {
    /* if not quoted, read up until the next seperator */

    while ((*e) && (!isspace(*e)) && (*e!='='))
      e++;

//    NEED TO SPEND MORE TIME THINKING ABOUT THIS!
// YAHOO is not relativizing right

    // skip over trailing spaces
    while ((*e) && (isspace(*e)))
      e++;

    // back up one if we're on a non-seperator
    if ((*e) && (!isspace(*e)) && (*e!='='))
      e--;
  }

  /* s = start of token, past any quotes */
  /* e = end of token, possible quote at *(e-1) */

  /* get rid of any leading spaces */
  while ((*s) && (isspace(*s)))
    s++;

  *last = *e;
  *s1 = (e+1);

  if (s!=e) {
    e--;
    while ((e!=s) && ((isspace(*e)) || (*e=='"')))
      e--;

    len = (e-s)+1;

    *dest = s;
    *(*dest+len)='\0';

    return TRUE;
  } else {
    *dest = NULL;
    return FALSE;
  }
}

void Parameterize(struct tnode *t, char *s)
{
 // struct tparam p;
  struct tparam *newparam;
  char last;
  char *ident, *data;

  last = *s;

  while (last) {
    if (PullData(&s,&ident,&last)) {
      if (last=='=')
        PullData(&s,&data,&last);
      else
        data=NULL;

      newparam = AllocateParamNode();
      newparam->Ident = ident;
      newparam->Data = data;

      newparam->Next = t->Params;
      t->Params = newparam;
    }
  }
}

struct tnode *ParseInit(struct tnode *headernode, struct Buffer *buf)
{
  assert(headernode!=NULL);

  if (!ModuleInit) {
    fprintf(stderr,"Nobody called ParseInitModule()\n");
    exit(1);
  }

  memset(headernode,0,sizeof(struct tnode));
  headernode->Tag=tagHeaderNode;

  Preform = FALSE;
  BufferGetChar(buf);

  LastToken = tagUnknown;

  if (dbuff==NULL) {
    dbuff = (char *) SafeMalloc(DBUFF_SIZE); /* 256k sounds about right */
    allocdbuff=TRUE;
  } else
    allocdbuff=FALSE;

  assert(dbuff!=NULL);

  return headernode;
}

void ParseCleanup(struct Buffer *buf)
{
  if ((dbuff!=NULL) && (allocdbuff)) {
    SafeFree(dbuff);
    dbuff=NULL;
    allocdbuff=FALSE;
  }
}

struct tnode *AddNode(struct tnode *parent, struct Buffer *buf)
{
  struct tnode *newnode = AllocateHtmlNode();

  assert(parent!=NULL);
  assert(newnode!=NULL);

  newnode->Tag = CurToken;
  newnode->Parent = parent;
  newnode->Next = NULL;
  newnode->Prev = NULL;
  newnode->Child = NULL;
  newnode->LastChild = NULL;
  newnode->Text = NULL;
  newnode->Params = NULL;

  if (dend!=dstart) {
    assert((dend-dstart)+1>0);
    newnode->Text = (char *) SafeMalloc((dend-dstart)+1);
    assert(newnode->Text!=NULL);
    memcpy(newnode->Text,dstart,((dend-dstart)+1));
  }

  if (parent->Child==NULL) {
    newnode->Prev = NULL;
    parent->Child = newnode;
    parent->LastChild = newnode;
  } else {
    assert(parent->LastChild!=NULL);

    newnode->Prev = parent->LastChild;
    parent->LastChild->Next = newnode;
    parent->LastChild = newnode;
  }

  if ((CurToken!=tagText) && (newnode->Text!=NULL))
    Parameterize(newnode,newnode->Text);

  return newnode;
}

/* parse links and images */
int HtmlParseLinks(struct tnode **rootnode, struct Buffer *buf)
{
  struct tnode *CurNode, headernode;
  int errors = 0;

  CurNode = ParseInit(&headernode,buf);

  while (GetToken(buf)) {
    if (CurToken==tagPre)
      Preform = !CurTok_Negate;
    else if ((dend!=dstart) && (IsLinkTag[CurToken]))
      AddNode(CurNode,buf);
  }

  *rootnode = headernode.Child;

  ParseCleanup(buf);

  return (errors==0);
}

int HtmlParse(struct tnode **rootnode, struct Buffer *buf)
{
  struct tnode *CurNode, *newnode, headernode;
  int errors = 0;

  CurNode = ParseInit(&headernode,buf);

  while (GetToken(buf)) {
    if ((CurTok_Negate) && (IsSTag[CurToken])) {
      /* do nothing */
    } else if (CurTok_Negate) {
      if (CurNode->Parent==NULL) {
        /* extra / somewhere -- do not move up a node */
      } else if (CurToken!=CurNode->Tag) {
        struct tnode *backtrack = CurNode;

        /* token mismatch, backtrack? */
        errors++;

        while ((backtrack!=NULL) && (backtrack->Tag!=CurToken))
          backtrack = backtrack->Parent;

        if (backtrack!=NULL) {
          /* backtrack successful */
          assert(backtrack!=&headernode);
          CurNode = backtrack->Parent;
          assert(CurNode!=NULL);
        } else {
          /* backtrack failed -- do not move up a node */
        }
      } else {
        /* negation tag matches current parent */
        CurNode = CurNode->Parent;
      }

      if (CurToken == tagPre)
        Preform = FALSE;
    } else if ((CurToken == tagSpace) && (!IsLSTag[LastToken])) {
      /* do nothing */
    } else {
      newnode = AddNode(CurNode,buf);

      if (!IsSTag[CurToken])
        CurNode = newnode;

      if (CurToken==tagPre)
        Preform = TRUE;
    }

    LastToken = CurToken;
  }

  *rootnode = headernode.Child;

  ParseCleanup(buf);

  return (errors==0);
}

void DisposeParameters(struct tparam *p)
{

  while (p!=NULL) {
    struct tparam *next = p->Next;

    FreeParamNode(p);
    p=next;
  }
}

void DisposeText(struct tnode *n)
{
  if (n->Text!=NULL) {
    SafeFree(n->Text);
    n->Text = NULL;
  }
}

void HtmlWrite(struct tnode *n, int filedes, char *buf)
{
  char *newline = "\n";

  while (n!=NULL) {
    struct tparam *p = n->Params;
    char *tokname = Lookup[n->Tag];

    if (*tokname) {
      strcpy(buf,"<");
      strcat(buf,tokname);

      while (p!=NULL) {
        strcat(buf," ");

        if (p->Ident!=NULL)
          strcat(buf,p->Ident);

        if (p->Data!=NULL) {
          strcat(buf,"=\"");
          strcat(buf,p->Data);
          strcat(buf,"\"");
        }

        p=p->Next;
      }

      strcat(buf,">");

      write(filedes,buf,strlen(buf));
    }

    if ((n->Text!=NULL) && (n->Tag==tagText))
      write(filedes,n->Text,strlen(n->Text));

    if (n->Tag==tagSpace)
      write(filedes," ",1);

    if (n->Child!=NULL)
      HtmlWrite(n->Child,filedes,buf);

    if ((*tokname) && (!IsSTag[n->Tag])) {
      strcpy(buf,"</");
      strcat(buf,tokname);
      strcat(buf,">");
      write(filedes,buf,strlen(buf));
    }

    n=n->Next;
  }

  write(filedes,newline,strlen(newline));
}

void HtmlDispose(struct tnode *n)
{
  while (n!=NULL) {
    struct tnode *next = n->Next;

    DisposeParameters(n->Params);
    DisposeText(n);

    if (n->Child!=NULL)
      HtmlDispose(n->Child);

    FreeHtmlNode(n);

    n=next;
  }
}

void LookupMappingSort(int l, int r)
{
  int i=l,j=r;
  struct LookupMappingNode y, x;

  x=LookupMapping[(i+j)/2];

  do {
    while (strcmp(LookupMapping[i].Name,x.Name)<0) i++;
    while (strcmp(x.Name,LookupMapping[j].Name)<0) j--;

    if (i<=j) {
      y=LookupMapping[i];
      LookupMapping[i]=LookupMapping[j];
      LookupMapping[j]=y;
      i++;
      j--;
    }

  } while (i<=j);

  if (l<j) LookupMappingSort(l,j);
  if (i<r) LookupMappingSort(i,r);
}

void ParseInitModule(void)
{
  int i;

  if (ModuleInit)
    return;

  CREATE_MUTEX(MutexParamFreeList);
  CREATE_MUTEX(MutexHtmlFreeList);

  for (i=0; i<TokCount; i++) {
    char *s;

    LookupMapping[i].Index = i;
    LookupMapping[i].Name = SafeStrDup(Lookup[i]);
    for (s=LookupMapping[i].Name; *s; s++)
      *s=(char) tolower(*s);
  }
  LookupMappingSort(0,TokCount-1);

  for (i=0; i<256; i++)
    LookupCharMapping[i] = -1;

  for (i=0; i<TokCount; i++) {
    unsigned char ch = (unsigned char) *LookupMapping[i].Name;

    if (LookupCharMapping[ch]==-1)
      LookupCharMapping[ch] = i;
  }

  for (i=0; i<TokCount; i++) {
    IsSTag[i] = is_stag(i);
    IsLSTag[i] = is_lstag(i);
    IsLinkTag[i] = is_linktag(i);
  }

  ModuleInit = TRUE;
}

