#include "stdinc.h"
#include "thread.h"
#include "safeget.h"

#ifndef WINDOWS
#include <netinet/in.h>
#include <arpa/inet.h>

#ifndef NOTHREAD
#define USE_SAFE_GET
#endif
#endif

#ifdef USE_SAFE_GET
pthread_key_t KeyHostent, Keyntoa;
pthread_once_t OnceSafeGet = PTHREAD_ONCE_INIT;
#endif

MUTEX_TYPE MutexGet;

void ThreadSafeGetInit(void)
{
  CREATE_MUTEX(MutexGet);

#ifdef USE_SAFE_GET
  pthread_key_create(&KeyHostent,NULL);
  pthread_key_create(&Keyntoa,NULL);
#endif
}

struct hostent *gethostbyname_safe(const char *name)
{
  struct hostent *h,*h_get;

  #ifdef USE_SAFE_GET
  pthread_once(&OnceSafeGet,ThreadSafeGetInit);

  if ((h=pthread_getspecific(KeyHostent))==NULL) {
    h = (struct hostent*) malloc(sizeof(struct hostent));
    pthread_setspecific(KeyHostent,h);
  }

  MUTEX_LOCK(MutexGet);

  h_get=gethostbyname(name);

  if (h_get==NULL)
    h=NULL;
  else
    *h=*h_get;

  MUTEX_UNLOCK(MutexGet);

  return h;
  #else
  return gethostbyname(name);
  #endif
}

char *inet_ntoa_safe(struct in_addr in)
{
  char *h,*h_get;

  #ifdef USE_SAFE_GET
  pthread_once(&OnceSafeGet,ThreadSafeGetInit);

  if ((h=pthread_getspecific(Keyntoa))==NULL) {
    h = (char*) malloc(256);
    pthread_setspecific(Keyntoa,h);
  }

  MUTEX_LOCK(MutexGet);

  h_get=inet_ntoa(in);

  if (h_get==NULL)
    h=NULL;
  else
    strcpy(h,h_get);

  MUTEX_UNLOCK(MutexGet);

  return h;
  #else
  return inet_ntoa(in);
  #endif
}




