#include <stdio.h>
#include "screen.h"

int NCOL;
int NROW;

#ifdef __TURBOC__

void scr_clr()
{
  clreol();
}

void scr_cls()
{
  clrscr();
}

void MOVE(int y, int x)
{
  gotoxy(x,y);
}

void scr_termcap()
{
  NROW=24;
  NCOL=80;
}

void scr_gotorc(int r, int c)
{
  gotoxy(c,r);
}

#else

#include <curses.h>
#include <term.h>

static void k_flip()
{
  struct termios tstuff;
  extern int errno;

  if (tcgetattr(0,&tstuff) == -1)
    printf("ioctl get error <%d>\n",errno);

  tstuff.c_lflag ^= (ECHO | ICANON);
  tstuff.c_cc[VMIN] = 1;
  tstuff.c_cc[VTIME] = 0;
  if (tcsetattr(0,TCSANOW,&tstuff) == -1)
    printf("ioctl put error <%d>\n",errno);
}

/* a blind proc for tputs() to use */
static int putch(char c)
{
        putc(c,stdout);
}

int NROW;
int NCOL;
char *PAD_CHAR;
bool AUTOWRAP;

/* termcap stuff */
char    tspace[256];    /* Space for capability strings */
char    *aoftspace;     /* Address of tspace for relocation */

char    *AL;            /* P* Add new blank line */
char    *AL_PARM;       /* P* Add n new blank lines */
char    *BC;            /*    Back cursor */
char    *BT;            /* P  Back tab */
char    *CD;            /* P* Clear to end of display */
char    *CE;            /* P  Clear to end of line */
char    *CL;            /* P* Clear screen */
char    *CM;            /* PG Cursor motion */
char    *CS;            /* PG Change scrolling region (vt100) */
char    *xCR;           /* P  Carriage return */
char    *DC;            /* P* Delete character */
char    *DL;            /* P* Delete line sequence */
char    *DL_PARM;       /* P* Delete n lines */
char    *DM;            /*    Delete mode (enter)  */
char    *DO;            /*    Down line sequence */
char    *DOWN_PARM;     /*    Down n lines */
char    *ED;            /*    End delete mode */
char    *EI;            /*    End insert mode */
char    *F0,*F1,*F2,*F3,*F4,*F5,*F6,*F7,*F8,*F9; /* function keys */
char    *HO;            /*    Home cursor */
char    *IC;            /* P  Insert character */
char    *IM;            /*    Insert mode (give as ':im=:' if 'ic' */
char    *IP;            /* P* Insert pad after char ins'd using IM+IE */
char    *KD;            /*    Keypad down arrow */
char    *KE;            /*    Keypad don't xmit */
char    *KH;            /*    Keypad home key */
char    *KL;            /*    Keypad left arrow */
char    *KR;            /*    Keypad right arrow */
char    *KS;            /*    Keypad start xmitting */
char    *KU;            /*    Keypad up arrow */
char    *LEFT_PARM;     /*    Left n chars */
char    *LL;            /*    Quick to last line, column 0 */
char    *LL;            /*    Quick to last line, column 0 */
char    *ND;            /*    Non-destructive space */
char    *RIGHT_PARM;    /*    Right n spaces */
char    *xNL;           /*    Line feed (new line) */
char    PC;             /*    Pad character */
char    *RC;            /*    Restore cursor from last SC */
char    *SC;            /*    Save cursor */
char    *SE;            /*    Standout end (may leave space) */
char    *SF;            /* P  Scroll forwards */
char    *SO;            /*    Stand out begin (may leave space) */
char    *SR;            /* P  Scroll backwards */
char    *TA;            /* P  Tab (other than ^I or with padding) */
char    *TE;            /*    Terminal end sequence */
char    *TI;            /*    Terminal initial sequence */
char    *UP;            /*    Upline */
char    *UP_PARM;       /*    Up n lines */
char    *VB;            /*    Visible bell */
char    *VE;            /*    Visual end sequence */
char    *VS;            /*    Visual start sequence */
bool    AM;             /* Automatic margins */
bool    xBS;            /* Backspace works */
bool    CA;             /* Cursor addressible */
bool    DA;             /* Display may be retained above */
bool    DB;             /* Display may be retained below */
bool    EO;             /* Can erase overstrikes with ' ' */
bool    GT;             /* Gtty indicates tabs */
bool    HC;             /* Hard copy terminal */
bool    HZ;             /* Hazeltine ~ braindamage */
bool    IN;             /* Insert-null blessing */
bool    MI;             /* can move in insert mode */
bool    NC;             /* No Cr - \r snds \r\n then eats \n (dm2500) */
bool    NS;             /* No scroll - linefeed at bottom won't scroll */
bool    OS;             /* Overstrike works */
bool    UL;             /* Underlining works even though !os */
bool    XB;             /* Beehive (no escape key, simulate with f1) */
bool    XN;             /* A newline gets eaten after wrap (concept) */
bool    XT;             /* Tabs are destructive */
bool    XX;             /* Tektronix 4025 insert line */

/* holds termcap information */
char **sstrs[] = {
&AL,&BC,&BT,&CD,&CE,&CL,&CM,&xCR,&CS,&DC,&DL,&DM,&DO,&ED,&EI,&F0,
&F1,&F2,&F3,&F4,&F5,&F6,&F7,&F8,&F9,&HO,&IC,&IM,&IP,&KD,&KE,&KH,
&KL,&KR,&KS,&KU,&LL,&ND,&xNL,&PAD_CHAR,&RC,&SC,&SE,&SF,&SO,&SR,
&TA,&TE,&TI,&UP,&VB,&VS,&VE,&AL_PARM,&DL_PARM,&UP_PARM,&DOWN_PARM,
&LEFT_PARM,&RIGHT_PARM};

bool *sflags[] = {
&AUTOWRAP,&xBS,&DA,&DB,&EO,&HC,&HZ,&IN,&MI,&NC,&NS,&OS,&UL,&XB,&XN,&XT,&XX};

#define MOVE(row,col)   tputs(tgoto(CM,col-1,row-1),0,putch)


/* scr_clr - clear the current row */
void
scr_clr()
{
tputs(CE,1,putch);
}


/* scr_cls - clear the screen */
void
scr_cls()
{
tputs(CL,1,putch);
}


void s_errmsg(char *s, int n)
{
  printf("%s\n",s);
  exit();
}

/* termcap initialization routine */
void
scr_termcap()
{
  char *cp;
  char *getenv();
  char *tgetstr();
  char termbuf[1024];
  register char *namp;
  register bool **fp;
  register char ***sp;

  if ((cp = getenv("TERM")) == NULL || !*cp)
    cp = "dumb";
  if (tgetent(termbuf,cp) != 1)
    s_errmsg("Terminal type is unknown",0);

  namp = "ambsdadbeohchzinmincnsosulxbxnxtxx";
  fp = sflags;
  do {
    *(*fp++) = tgetflag(namp);
    namp += 2;
  } while (*namp);
  aoftspace = tspace;
  namp = "albcbtcdceclcmcrcsdcdldmdoedeik0k1k2k3k4k5k6k7k8k9hoicimipkdkekh
  klkrkskullndnlpcrcscsesfsosrtatetiupvbvsveALDLUPDOLERI";
  sp = sstrs;
  do {
    *(*sp++) = tgetstr(namp,&aoftspace);
    namp += 2;
  } while (*namp);
  NROW = tgetnum("li");
  NCOL = tgetnum("co");
  AUTOWRAP = 1;
  if (PAD_CHAR == NULL) {
    PC = '\0';
  PAD_CHAR = &PC;
  }
}

void scr_gotorc(int r, int c)
{
  MOVE(r,c);
}

#endif


