#include "stdinc.h"
#include <signal.h>
#include <assert.h>
#include <stdio.h>
#include <rpc/rpc.h>
#include "nfs.h"
#include "thread.h"
#include "servmain.h"
#include "preferences.h"
#include "thread.h"
#include "myrun.h"

#include "stats.h"


void ServerInit(void);

extern SOCKET SocketForHttp;


// I include the source of tnfs_svc.c inline below. I'm not sure really why I

// chose to do it this way, probably had something to do with rpcgen creating

// nfs_svc automatically.

#include "tnfs_svc.c"


struct sPreferences Preferences;

THREAD_TYPE ThreadMain;

int MainThreadInSelect;

static struct { char *name, *use ; } Args[] = {
  {"-h",          	"Print usage information"},
  {"-webdir",     	"Web cache directory name (default is -webdir ./webcache)"},
  {"-weblevels",  	"Web cache directory levels (default is -weblevels 1)"},
  {"-weblevwidth",	"Web cache directory level width (default is -weblevelwidth 1)"},
  {"-mountport", 	"Specify mount port "},
  {"-nfsport",   	"Specify nfs port"},
  {"-doshutdown", 	"Allow use of shutdown() socket call"},
  {"-proxyname",	"Specify proxy name (default is no proxy}"},
  {"-proxyport",	"Specify proxy port (default -proxyport 80}"},
  {"-noregister",	"Do not register NFS/Mount services"},
  {"-maxhandlecount",	"Specify maximum NFS handle count"},
  {"-debug",		"Enable debug messages"},
  {"-detach",		"Detach server from terminal"},
  {"-exec",		"Make .contents files executable"},
  {"-workers",		"Number of worker threads"},
  {"-linkimages",	"Create .image files for images"},
  {"-hist",		"Produce histogram output"},
  {"-score",		"Display scoreboard"},
  {"-local",		"Restrict requests to local net"},
  {NULL, NULL}
  } ;

static void usage( char *name ) {
  int f ;

  printf( "usage: %s [flags]\n", name ) ;
  printf( "The following flags are recognized:\n" ) ;
  for( f = 0 ; Args[f].name != NULL ; f++ )
    printf( "flag: %-5s  meaning: %s\n", Args[f].name, Args[f].use ) ;
  }

static void ProcessCommandLine( int argc, char *argv[] ) {
  int a, f ;

  strcpy(Preferences.TempPath,"./webcache");
  Preferences.NumRestrict=0;
  Preferences.Executable = FALSE;
  Preferences.DirLevels = 1;
  Preferences.DirLevelWidth = 1;
  Preferences.MountPort = -1;
  Preferences.NfsPort = -1;
  Preferences.RegisterServices = TRUE;
  Preferences.MaxOpenFiles = 64;
  Preferences.MaxDiskFiles = 4096; //256;
  Preferences.AllowShutdown = FALSE; // breaks www.espn.com -- TRUE;
  *Preferences.ProxyName = '\0';
  Preferences.ProxyPort = 80;
  Preferences.MaxHandleCount = 16384; //64L*1024L;
  Preferences.Debug = FALSE;
  Preferences.Detach = FALSE;
  Preferences.WorkerThreads = 8;
  Preferences.LinkImages = FALSE;
  Preferences.Histograms = FALSE;
  Preferences.Statistics = FALSE;
  Preferences.ScoreBoard = FALSE;

  for( a = 1 ; a < argc ; a++ ) {
    char *cmdname = argv[a];

    for( f = 0 ; Args[f].name != NULL ; f++ )
      if( stricmp(Args[f].name, cmdname) == 0 ) break ;

    switch( f ) {
      case 0: usage("server");
      	      exit(1);
              break; /* usage information */

      case 1: if ((a+1<argc) && (*argv[++a]!='\0'))
                strcpy(Preferences.TempPath,argv[a]);
	      break;

      case 2: if ((a+1<argc) && (*argv[++a]!='\0'))
                Preferences.DirLevels = atoi(argv[a]);
	      break;

      case 3: if ((a+1<argc) && (*argv[++a]!='\0'))
                Preferences.DirLevelWidth = atoi(argv[a]);
      	      break;

      case 4: if ((a+1<argc) && (*argv[++a]!='\0'))
                Preferences.MountPort = atoi(argv[a]);
      	      break;

      case 5: if ((a+1<argc) && (*argv[++a]!='\0'))
                Preferences.NfsPort = atoi(argv[a]);
              break;

      case 6: Preferences.AllowShutdown = TRUE;
      	      break;

      case 7: if ((a+1<argc) && (*argv[++a]!='\0'))
                strcpy(Preferences.ProxyName,argv[a]);
      	      break;

      case 8: if ((a+1<argc) && (*argv[++a]!='\0'))
      		Preferences.ProxyPort = atoi(argv[a]);
              break;

      case 9: Preferences.RegisterServices = FALSE;
      	      break;

      case 10: if ((a+1<argc) && (*argv[++a]!='\0'))
                 Preferences.MaxHandleCount = atoi(argv[a]);
               break;

      case 11: Preferences.Debug=TRUE;
      	       break;

      case 12: Preferences.Detach=TRUE;
      	       break;

      case 13: Preferences.Executable=TRUE;
      	       break;

      case 14: if ((a+1<argc) && (*argv[++a]!='\0'))
                 Preferences.WorkerThreads = atoi(argv[a]);
      	       break;

      case 15: Preferences.LinkImages = TRUE;
      	       break;

      case 16: Preferences.Histograms = TRUE;
      	       Preferences.Statistics=TRUE;
      	       break;

      case 17: Preferences.ScoreBoard=TRUE;
      	       Preferences.Statistics=TRUE;
      	       break;

      case 18: Preferences.NumRestrict=1;
      	       Preferences.RestrictNet[0] = inet_addr("192.12.69.0");
               Preferences.RestrictSubNet[0] = inet_addr("255.255.255.0");
               break;
    }
  }
}

void Detach(void)
{
  int pid = fork();
  int i;

  if (pid<0) {
    perror("unable to fork!");
    exit(1);
  }

  if (pid)
    exit(0);

  setsid();

  fprintf(stderr,"\nServer Detached form terminal.\n");

  i=open("/dev/null",2);
  dup2(i,1);
  dup2(i,2);
}

void InitRpc(void)
{
        SVCXPRT *transp = ((Preferences.NfsPort==-1) || (Preferences.MountPort==-1)) ? svcudp_create(RPC_ANYSOCK) : NULL;
        SOCKET nfssocket = (Preferences.NfsPort==-1) ? INVALID_SOCKET : socket(AF_INET,SOCK_DGRAM,0);
        SOCKET mountsocket = (Preferences.MountPort==-1) ? INVALID_SOCKET : socket(AF_INET,SOCK_DGRAM,0);
        SVCXPRT *nfstransp;
        SVCXPRT *mounttransp;
        int mountproto = (Preferences.MountPort==-1) ? IPPROTO_UDP : 0;
        int nfsproto = (Preferences.NfsPort==-1) ? IPPROTO_UDP : 0;
        struct sockaddr_in inaddr;

        inaddr.sin_family = AF_INET;
        inaddr.sin_addr.s_addr = INADDR_ANY;

        if (Preferences.NfsPort!=-1) {
          inaddr.sin_port = htons(Preferences.NfsPort);

          if (nfssocket==INVALID_SOCKET) {
            fprintf(stderr,"unable to create nfs udp socket\n");
            return;
          } else if (bind(nfssocket,(struct sockaddr*) &inaddr,sizeof(inaddr))!=0) {
            fprintf(stderr,"unable to bind nfs udp socket\n");
            return;
          }
        }

        if (Preferences.MountPort!=-1) {
          inaddr.sin_port = htons(Preferences.MountPort);

          if (mountsocket==INVALID_SOCKET) {
            fprintf(stderr,"unable to create mount udp socket\n");
            return;
          } else if (bind(mountsocket,(struct sockaddr*) &inaddr,sizeof(inaddr))!=0) {
            fprintf(stderr,"unable to bind mount udp socket\n");
            return;
          }
        }

        nfstransp = (Preferences.NfsPort==-1) ? transp : svcudp_create(nfssocket);
        mounttransp = (Preferences.MountPort==-1) ? transp : svcudp_create(mountsocket);

        if (Preferences.RegisterServices) {
      	  (void) pmap_unset(NFS_PROGRAM, NFS_VERSION);
          (void) pmap_unset(MOUNT_PROGRAM, MOUNTVERS);

          if (!svc_register(nfstransp, MOUNT_PROGRAM, MOUNTVERS, mount_program_1, mountproto)) {
                  fprintf(stderr, "unable to register (MOUNT_PROGRAM, MOUNTVERS, udp).");
                  exit(1);
          }
          if (!svc_register(mounttransp, NFS_PROGRAM, NFS_VERSION, nfs_program_2, nfsproto)) {
                  fprintf(stderr, "unable to register (NFS_PROGRAM, NFS_VERSION, udp).");
                  exit(1);
          }
        }

        /*
  	transp = svctcp_create(RPC_ANYSOCK, 0, 0);
	if (transp == NULL) {
		fprintf(stderr, "cannot create tcp service.");
		exit(1);
	}
    	if (!svc_register(transp, MOUNT_PROGRAM, MOUNTVERS, mount_program_1, IPPROTO_TCP)) {
		fprintf(stderr, "unable to register (MOUNT_PROGRAM, MOUNTVERS, tcp).");
		exit(1);
	}
	if (!svc_register(transp, NFS_PROGRAM, NFS_VERSION, nfs_program_2, IPPROTO_TCP)) {
		fprintf(stderr, "unable to register (NFS_PROGRAM, NFS_VERSION, tcp).");
		exit(1);
	}
        */

        ServerInit();

	my_svc_run(SocketForHttp);
}

int
main(int argc, char **argv)
{
	register SVCXPRT *transp;

#ifdef NEED_PTHREAD_INIT
	pthread_init();
#endif
        scr_termcap();

        ProcessCommandLine(argc,argv);

        InitStatistics();

        if (Preferences.Detach)
  	  Detach();

        CREATE_MUTEX(MutexServer);
        signal(SIGPIPE,SIG_IGN);
        CreateThreadQueue();
        CreateWorkerThreads();

        ThreadMain = pthread_self();
        MainThreadInSelect = FALSE;

        InitRpc();

	fprintf(stderr, "svc_run returned");
	exit(1);
}
