#ifndef smalloc_h
#define smalloc_h

#ifdef __CPLUSPLUS__
extern "C" {
#endif

#ifdef USE_SAFE_MALLOC
#define MALLOC(x) SafeMalloc1((x),__FILE__,__LINE__)
#define FREE(x) SafeFree(x)
#define STRDUP(x) SafeStrDup1((x),__FILE__,__LINE__)
#else
#define MALLOC(x) malloc(x)
#define FREE(x) free(x)
#define STRDUP(x) strdup(x)
#endif

#define CHECKFREE(x) if (x!=NULL) { FREE(x); x=NULL; }
#define STRREPLACE(dest,src) { if (dest!=NULL) FREE(dest); dest=STRDUP(src); }
#define STRDUP_COND(x) (((x)==NULL) ? NULL : STRDUP((x)))

#ifdef SAFE_MALLOC_TRACE
#define SAFE_MALLOC_MUTEX
#endif

void SafeMallocDump(void);
void SafeMallocInit(void);
void *SafeMalloc1(int size, char *fn, int line);
void *SafeMalloc(int size);
void SafeFree(void *x);
char *SafeStrDup1(char *x, char *fn, int line);
char *SafeStrDup(char *x);

#ifdef __CPLUSPLUS__
}
#endif

#endif
