#include "stdinc.h"
#include <assert.h>
//#include <curses.h>
#include "stats.h"
#include "thread.h"
#include "screen.h"
#include "preferences.h"

#define MAX_STAT_DEPTH 16
#define MAX_STAT_COUNT 256

extern int NumberOfNodesDeleted, NumberOfNodesCreated, NumberOfNodes;

//#define MOVE(row,col)   tputs(tgoto(CM,col-1,row-1),0,putch)

int StatInfoCount=0;
struct StatInfo Stats[MAX_STAT_COUNT];

pStatRecord FirstStatistic = NULL;
int StatCount = 0;
int StatInit = FALSE;
TIMESTAMP LastStatisticsTime;
MUTEX_TYPE MutexStatistics;

void UpdateNodeCount()
{
  static int lastc, lastd, last;

  if ((lastc!=NumberOfNodesCreated) || (lastd!=NumberOfNodesDeleted) || (last!=NumberOfNodes)) {
    lastc=NumberOfNodesCreated;
    lastd=NumberOfNodesDeleted;
    last=NumberOfNodes;

    scr_gotorc(19,1);
    fprintf(stdout,"%d created, %d deleted, %d total",lastc,lastd,last);
  }
}

void MsgArea()
{
  if (!Preferences.ScoreBoard)
    return;

  scr_gotorc(20,1);
}

void InitStatistics()
{
  CREATE_MUTEX(MutexStatistics);
  SET_TIMESTAMP(LastStatisticsTime);

  memset(Stats,0,sizeof(Stats));

  if (Preferences.ScoreBoard) {
    scr_cls();
    MsgArea();
  }
}

pStatInfo FindStat(pid_t pid)
{
  pStatInfo stat;
  int i;
  // this should be optimized with thread local storage of stat

  for (i=0; i<StatInfoCount; i++)
    if (pid==Stats[i].Pid) {
      return &Stats[i];
    }

  assert(StatInfoCount<MAX_STAT_COUNT);
  stat=&Stats[StatInfoCount];
  StatInfoCount++;
  stat->Pid=pid;
  stat->Index=StatInfoCount-1;

  return stat;
}

void UpdateLine(pStatInfo stat)
{
  if (!Preferences.ScoreBoard)
    return;

  if ((stat->DispCount%100)==0) {
    scr_gotorc(1,1);
    fprintf(stdout,"# PID");
    scr_gotorc(1,25);
    fprintf(stdout,"D %-25s %-25s","CURRENT","LAST");
  }

  if ((stat->DispCount++)%10==0) {
    scr_gotorc(stat->Index+2,1);
    fprintf(stdout,"%d %d",stat->Index,stat->Pid);
    scr_clr();
  }

  scr_gotorc(stat->Index+2,25);

  fprintf(stdout,"%d %-25s %-25s",
    stat->Depth,
    (stat->Depth==0 ? "(none)" : stat->Names[stat->Depth-1]),
    (stat->Names[stat->Depth]==NULL ? "(none)" : stat->Names[stat->Depth]));

  UpdateNodeCount();
  MsgArea();
  fflush(stdout);
}

pStatInfo StartStatistics(char *statname)
{
  pStatInfo stat;
  char **name;
  TIMESTAMP *time;

  if (!Preferences.Statistics)
    return NULL;

  MUTEX_LOCK(MutexStatistics);

  stat = FindStat(getpid());

  assert(stat->Depth<MAX_STAT_DEPTH);
  name = &(stat->Names[stat->Depth]);
  time = &(stat->Times[stat->Depth]);
  stat->Depth++;

  assert(strlen(statname)<256);
  if (*name==NULL)
    *name = (char*) malloc(256);

  strcpy(*name,statname);
  SET_TIMESTAMP(*time);

  UpdateLine(stat);

  MUTEX_UNLOCK(MutexStatistics);

  return stat;
}

void StatShowURL(char *x)
{
  if (x==NULL)
    return;

  if (!Preferences.ScoreBoard)
    return;

  scr_gotorc(18,1);
  fprintf(stdout,"%s",x);
  scr_clr();
  fflush(stdout);
}

void RecordStatistics(char *name, pStatInfo statinfo)
{
  pStatRecord stat;
  TIMESTAMP last;
  TIMESTAMP first;
  int elapsed,hbig,hsmall,elap10;

  if (!Preferences.Statistics)
    return;

  MUTEX_LOCK(MutexStatistics);

  assert(statinfo!=NULL);

  if (statinfo->Depth<=0) {
    fprintf(stderr,"%d %s ",statinfo->Depth,name);
    assert(statinfo->Depth>0);
  }

  statinfo->Depth--;
  name = statinfo->Names[statinfo->Depth];
  first = statinfo->Times[statinfo->Depth];

  SET_TIMESTAMP(last);
  elap10 = GET_ELAPSED10(first,last);
  elapsed = elap10/10;

  if (Preferences.Histograms) {
    stat = FirstStatistic;

    while ((stat!=NULL) && (stat->Name!=name))
      stat = stat->Next;

    if (stat==NULL) {
      stat = (pStatRecord) malloc(sizeof(struct StatRecord));
      memset(stat,0,sizeof(struct StatRecord));
      stat->Next = FirstStatistic;
      FirstStatistic=stat;
      stat->Name = name;
    }

    hsmall=elapsed;
    if (hsmall>=MAX_STAT_ELAPSED)
      hsmall = MAX_STAT_ELAPSED-1;

    stat->HistoSmall[hsmall]++;

    hbig=elapsed/MULTIPLIER;
    if (hbig>=MAX_STAT_ELAPSED)
      hbig = MAX_STAT_ELAPSED-1;

    stat->HistoBig[hbig]++;

    stat->Total++;
    stat->TotalTime+=elap10;

    StatCount++;
    if (StatCount>STAT_COUNT_LIMIT) {
      DumpStatistics();
      StatCount=0;
    }
  }

  UpdateLine(statinfo);

  MUTEX_UNLOCK(MutexStatistics);
}

void DumpStatistics()
{
  int i;
  FILE *f = fopen("stats.txt","at");
  time_t t;
  TIMESTAMP tstamp;
  pStatRecord stat = FirstStatistic;

  if (f==NULL)
    return;

  SET_TIMESTAMP(tstamp);

  time(&t);
  fprintf(f,"----- %d %s",GET_ELAPSED(LastStatisticsTime,tstamp),ctime(&t));

  LastStatisticsTime = tstamp;

  while (stat!=NULL) {
    if (stat->Total>0) {
      fprintf(f,"%s: %d %d ",stat->Name,stat->Total,stat->TotalTime/10);

      for (i=0; i<10; i++)
        if (stat->HistoSmall[i]>0) {
          fprintf(f,"%d/%d ",stat->HistoSmall[i],i);
          stat->HistoSmall[i]=0;
        }

      fprintf(f,"* %d ",MULTIPLIER);

      for (i=0; i<MAX_STAT_ELAPSED; i++)
        if (stat->HistoBig[i]>0) {
          fprintf(f,"%d/%d ",stat->HistoBig[i],i*MULTIPLIER);
          stat->HistoBig[i]=0;
        }

      fprintf(f,"\n");
      stat->Total=0;
      stat->TotalTime=0;
    }

    stat=stat->Next;
  }

  fclose(f);
}

