#ifndef THREAD_H
#define THREAD_H

#ifdef NOTHREAD
#define CREATE_SEMAPHORE(x) x=0
#define WAIT_SEMAPHORE(x) x=0
#define SIGNAL_SEMAPHORE(x) x=0
#define SEMAPHORE_TYPE int
#define THREAD_TYPE int
#define MUTEX_TYPE int
#define CREATE_MUTEX(x) x=0
#define MUTEX_LOCK(x) x=0
#define MUTEX_UNLOCK(x) x=0
#define CREATE_THREAD(x,start,param) fubar
#define MAKEDIR(x) mkdir(x,0777)
#else

#ifdef WINDOWS
#define CREATE_SEMAPHORE(x) x = CreateSemaphore(NULL,0,1024,NULL)
#define WAIT_SEMAPHORE(x) WaitForSingleObject(x,INFINITE)
#define WAIT_SEMAPHORE_TIMEOUT(x,t) WaitForSingleObject(x,t)
#define SIGNAL_SEMAPHORE(x) ReleaseSemaphore(x,1,NULL)
#define SEMAPHORE_TYPE HANDLE
#define THREAD_TYPE unsigned long
#define MUTEX_TYPE HANDLE
#define CREATE_MUTEX(x) x=CreateMutex(NULL,FALSE,NULL)
#define MUTEX_LOCK(x) WaitForSingleObject(x,INFINITE)
#define MUTEX_UNLOCK(x) ReleaseMutex(x)
#define CREATE_THREAD(x,start,param) x=_beginthread(start,0,param)
#define MAKEDIR(x) mkdir(x)

#define COND_TYPE HANDLE
#define COND_BROADCAST(x) PulseEvent(x);
//#define COND_SIGNAL(x) PulseEvent(x);
#define COND_WAIT(cond,mutex) {MUTEX_UNLOCK(mutex); WaitForSingleObject(cond,INFINITE); MUTEX_LOCK(mutex);}
#define COND_INIT(x) x=CreateEvent(NULL,TRUE,FALSE,NULL);

#else

#ifdef USE_GTHREADS
#include <gthreads.h>
#else
#include <pthread.h>
//#include <pthread/mit/pthread.h>
#endif

/* unix! */
#define THREAD_TYPE pthread_t
#define MUTEX_TYPE pthread_mutex_t
#define CREATE_MUTEX(x) pthread_mutex_init(&x,NULL)
#define MUTEX_LOCK(x) pthread_mutex_lock(&x)
#define MUTEX_UNLOCK(x) pthread_mutex_unlock(&x)
#define CREATE_THREAD(x,start,param) pthread_create(&x,NULL,start,param)

#define COND_TYPE pthread_cond_t
#define COND_INIT(x) pthread_cond_init(&x,NULL)
#define COND_WAIT(cond,mutex) pthread_cond_wait(&cond,&mutex)
#define COND_SIGNAL(cond) pthread_cond_signal(&cond)
#define COND_BROADCAST(cond) pthread_cond_broadcast(&cond);

#define MAKEDIR(x) mkdir(x,S_IRWXU)

/*
#define SEMAPHORE_TYPE sem_t
#define CREATE_SEMAPHORE(x) sem_init(&x,0,0)
#define WAIT_SEMAPHORE(x) sem_wait(&x)
#define SIGNAL_SEMAPHORE(x) sem_port(&x)
*/

/* implement semaphores using monitors (Dr. Andrews' book, pg 310) */
struct my_semaphore {
  MUTEX_TYPE Mutex;
  COND_TYPE Condition;
  int Count;
};

#define SEMAPHORE_TYPE struct my_semaphore
#define WAIT_SEMAPHORE(x) sem_wait(&x)
#define CREATE_SEMAPHORE(x) sem_init(&x)
#define SIGNAL_SEMAPHORE(x) sem_post(&x)

#endif
#endif

#define MUTEX_INIT(x) CREATE_MUTEX(x)

#endif
