#define THREAD_QUEUE_SIZE 256

pthread_key_t KeyThreadParams;
int InProgressCount=0;

/*
 * kept in xprt->xp_p2
 * this is private data held by 'svc_udp.c'. It is implementation specific and
 *   probably not protable.
 */
struct svcudp_data {
	u_int   su_iosz;	/* byte size of send.recv buffer */
	u_long	su_xid;		/* transaction id */
	XDR	su_xdrs;	/* XDR handle */
	char	su_verfbody[MAX_AUTH_BYTES];	/* verifier body */
	char * 	su_cache;	/* cached data, NULL if no cache */
};
#define	su_data(xprt)	((struct svcudp_data *)(xprt->xp_p2))

union uargument {
        dirpath mountproc_mnt_1_arg;
        dirpath mountproc_umnt_1_arg;
        fhandleargs nfsproc_getattr_2_arg;
        sattrargs nfsproc_setattr_2_arg;
        diropargs nfsproc_lookup_2_arg;
        fhandleargs nfsproc_readlink_2_arg;
        readargs nfsproc_read_2_arg;
        writeargs nfsproc_write_2_arg;
        createargs nfsproc_create_2_arg;
        diropargs nfsproc_remove_2_arg;
        renameargs nfsproc_rename_2_arg;
        linkargs nfsproc_link_2_arg;
        symlinkargs nfsproc_symlink_2_arg;
        createargs nfsproc_mkdir_2_arg;
        diropargs nfsproc_rmdir_2_arg;
        readdirargs nfsproc_readdir_2_arg;
        fhandleargs nfsproc_statfs_2_arg;
};

// maximum size of an NFS result. The result buffer is statically allocated,
// one per thread via thread local storage. The number is so big becuase Gecko2
// puts it's read bytes in the result (as opposed to gecko1, which used mmap'd
// io

#define RESULTSPACE_SIZE (16384+256)

typedef struct threadparams *pThreadParams;

struct threadparams {
  union uargument argument;
  struct svc_req rqstp;
  bool_t (*xdr_argument)(), (*xdr_result)();
  char *(*local)();
  SVCXPRT transp;
  pThreadParams Next;
  pThreadParams NextInProgress,PrevInProgress; // used only by InProgressQueue
  char *result;
  char ResultSpace[RESULTSPACE_SIZE];          // bytes to hold NFS results
  struct svcudp_data udpdata;
  int pid;
};


MUTEX_TYPE MutexServer, MutexRPC;

SEMAPHORE_TYPE SemaphoreRequestWaiting;
MUTEX_TYPE MutexThreadQueue;
pThreadParams ThreadQueue, ThreadQueueFree, ThreadQueueLast, ThreadReplyQueue, ThreadReplyQueueLast;
pThreadParams InProgressQueue; // not really a queue

void *AllocResult(int bytes)
{
  pThreadParams params = (pThreadParams) pthread_getspecific(KeyThreadParams);

  assert(bytes<RESULTSPACE_SIZE);

  return (void*) &params->ResultSpace;
}

void AddInProgress(pThreadParams newnode)
{
  assert(newnode!=NULL);

  MUTEX_LOCK(MutexThreadQueue);

  newnode->NextInProgress = InProgressQueue->NextInProgress;
  newnode->PrevInProgress = InProgressQueue;
  InProgressQueue->NextInProgress->PrevInProgress = newnode;
  InProgressQueue->NextInProgress = newnode;

  InProgressCount++;

  MUTEX_UNLOCK(MutexThreadQueue);
}

void RemoveInProgress(pThreadParams newnode)
{
  assert(newnode!=NULL);

  MUTEX_LOCK(MutexThreadQueue);

  newnode->PrevInProgress->NextInProgress=newnode->NextInProgress;
  newnode->NextInProgress->PrevInProgress=newnode->PrevInProgress;

  InProgressCount--;

  MUTEX_UNLOCK(MutexThreadQueue);
}

int TestInProgress(pThreadParams x)
{
  pThreadParams cur;
  char tmp[256];

#define EQADDR(a1, a2)   (bcmp((char*)&a1, (char*)&a2, sizeof(a1)) == 0)

  MUTEX_LOCK(MutexThreadQueue);

  for (cur=InProgressQueue->NextInProgress; cur!=InProgressQueue; cur=cur->NextInProgress)
    if ((cur->rqstp.rq_prog==x->rqstp.rq_prog) &&
       (cur->rqstp.rq_vers==x->rqstp.rq_vers) &&
       (cur->rqstp.rq_proc==x->rqstp.rq_proc) &&
       (cur->udpdata.su_xid==x->udpdata.su_xid) &&
       (EQADDR(cur->transp.xp_raddr,x->transp.xp_raddr))) {

      sprintf(tmp,"owner's pid = %d semcnt=%d ipcount=%d",cur->pid,SemaphoreRequestWaiting.Count,InProgressCount);
      debug("In Progress",tmp,NULL);

      MUTEX_UNLOCK(MutexThreadQueue);

      return TRUE;
    }

  MUTEX_UNLOCK(MutexThreadQueue);
  return FALSE;
}

pThreadParams GetThreadQueue(void)
{
  pThreadParams newnode;

  MUTEX_LOCK(MutexThreadQueue);

  newnode = ThreadQueue;
  if (newnode!=NULL)
    ThreadQueue=ThreadQueue->Next;

  MUTEX_UNLOCK(MutexThreadQueue);

  return newnode;
}

void AddThreadQueue(pThreadParams node)
{
  MUTEX_LOCK(MutexThreadQueue);

  node->Next = NULL;

  if (ThreadQueue!=NULL) {
    ThreadQueueLast->Next = node;
    ThreadQueueLast = node;
  } else {
    ThreadQueue = node;
    ThreadQueueLast = node;
  }

  MUTEX_UNLOCK(MutexThreadQueue);
}

pThreadParams AllocThreadQueue(void)
{
  pThreadParams result;

  MUTEX_LOCK(MutexThreadQueue);

  assert(ThreadQueueFree!=NULL);

  result = ThreadQueueFree;
  ThreadQueueFree = ThreadQueueFree->Next;

  MUTEX_UNLOCK(MutexThreadQueue);

  return result;
}

void FreeThreadQueue(pThreadParams node)
{
  MUTEX_LOCK(MutexThreadQueue);

  node->Next = ThreadQueueFree;
  ThreadQueueFree = node;

  MUTEX_UNLOCK(MutexThreadQueue);
}

void CreateThreadQueue(void)
{
  int i;

  CREATE_MUTEX(MutexRPC);
  CREATE_MUTEX(MutexThreadQueue);

  ThreadQueue = NULL;
  ThreadQueueFree = NULL;
  ThreadReplyQueue = NULL;

  for (i=0; i<THREAD_QUEUE_SIZE; i++) {
    pThreadParams newnode = (pThreadParams) malloc(sizeof(struct threadparams));

    newnode->Next = ThreadQueueFree;
    ThreadQueueFree = newnode;
  }

  // InProgressQueue has a header node
  InProgressQueue=(pThreadParams) malloc(sizeof(struct threadparams));
  InProgressQueue->NextInProgress = InProgressQueue;
  InProgressQueue->PrevInProgress = InProgressQueue;
}

void ProcessReply(pThreadParams tp)
{
  MUTEX_LOCK(MutexRPC);

  /* the next line is probably NOT PORTABLE */
  su_data((&tp->transp))->su_xid = tp->udpdata.su_xid;

  if (tp->result != NULL && !svc_sendreply(&tp->transp, tp->xdr_result, tp->result)) {
          svcerr_systemerr(&tp->transp);
  }

  if (!svc_freeargs(&tp->transp, tp->xdr_argument, (caddr_t) (&tp->argument))) {
          fprintf(stderr, "unable to free arguments");
          exit(1);
        }

//  FreeResult(tp->result);

  MUTEX_UNLOCK(MutexRPC);

  RemoveInProgress(tp);
  FreeThreadQueue(tp);
}

void ProcessRequest(pThreadParams tp)
{
  pStatInfo stats;

  pthread_setspecific(KeyThreadParams,(void*) tp);

  stats=StartStatistics("monitorwait");
  MUTEX_LOCK(MutexServer);
  RecordStatistics("monitorwait",stats);
  stats=StartStatistics("work");
  tp->result = (*(tp->local))((char *) (&(tp->argument)), &tp->rqstp);
  RecordStatistics("work",stats);
  MUTEX_UNLOCK(MutexServer);

  stats=StartStatistics("reply");
  ProcessReply(tp);
  RecordStatistics("reply",stats);
}

void *ThreadRPC(void *arg)
{
  while (TRUE) {
    pThreadParams tp;
    pStatInfo stats;

    stats = StartStatistics("waiting");

    WAIT_SEMAPHORE(SemaphoreRequestWaiting);

    RecordStatistics("waiting",stats);

    tp=GetThreadQueue();
    assert(tp!=NULL);

    tp->pid = getpid();

    ProcessRequest(tp);
  }
}

void CreateWorkerThreads()
{
  int i;
  THREAD_TYPE handle;

  CREATE_SEMAPHORE(SemaphoreRequestWaiting);

  for (i=0; i<Preferences.WorkerThreads; i++)
    CREATE_THREAD(handle,ThreadRPC,NULL);

  pthread_key_create(&KeyThreadParams,NULL);
}

int CheckNet(SVCXPRT *transp)
{
  struct sockaddr_in *caller_addr = svc_getcaller(transp);
  int i;

  if (Preferences.NumRestrict>0) {
    for (i=0; i<Preferences.NumRestrict; i++) {
      int caller_net = (caller_addr->sin_addr.s_addr & Preferences.RestrictSubNet[i]);
      int restrict_net = (Preferences.RestrictNet[i] & Preferences.RestrictSubNet[i]);

      if (caller_net == restrict_net)
        return TRUE;
    }
    return FALSE;
  } else
    return TRUE;
}

void
mount_program_1(rqstp, transp)
	struct svc_req *rqstp;
	register SVCXPRT *transp;
{
        pThreadParams tp;
        THREAD_TYPE handle;
        pStatInfo stats;

        if (!CheckNet(transp))
          return;

        stats=StartStatistics("mountproc");

        tp=AllocThreadQueue();

        tp->rqstp = *rqstp;
        tp->transp = *transp;

	switch (rqstp->rq_proc) {
	case MOUNTPROC_NULL:
		tp->xdr_argument = xdr_void;
		tp->xdr_result = xdr_void;
		tp->local =  (char *(*)()) mountproc_null_1;
		break;

	case MOUNTPROC_MNT:
		tp->xdr_argument = xdr_dirpath;
		tp->xdr_result = xdr_fhstatus;
		tp->local =  (char *(*)()) mountproc_mnt_1;
		break;

	case MOUNTPROC_DUMP:
		tp->xdr_argument = xdr_void;
		tp->xdr_result = xdr_mountlist;
		tp->local =  (char *(*)()) mountproc_dump_1;
		break;

	case MOUNTPROC_UMNT:
		tp->xdr_argument = xdr_dirpath;
		tp->xdr_result = xdr_void;
		tp->local =  (char *(*)()) mountproc_umnt_1;
		break;

	case MOUNTPROC_EXPORT:
		tp->xdr_argument = xdr_void;
		tp->xdr_result = xdr_exports;
		tp->local =  (char *(*)()) mountproc_export_1;
		break;

	default:
		svcerr_noproc(transp);
                FreeThreadQueue(tp);
                RecordStatistics("mountproc_noproc",stats);
		return;
	}
	bzero((char *)&(tp->argument), sizeof(tp->argument));

	if (!svc_getargs(transp, tp->xdr_argument, &tp->argument)) {
		svcerr_decode(transp);
                FreeThreadQueue(tp);
                RecordStatistics("mountproc_badargs",stats);
		return;
	}

        /* the next line is probably NOT PORTABLE */
        tp->udpdata = *(su_data((&tp->transp)));
        tp->pid = getpid();

        if (TestInProgress(tp)) {
          FreeThreadQueue(tp);
          RecordStatistics("mountproc_inprogress",stats);
        } else {
          AddInProgress(tp);
  	  AddThreadQueue(tp);
          SIGNAL_SEMAPHORE(SemaphoreRequestWaiting);
          RecordStatistics("mountproc_signal",stats);
        }

	return;
}

void
nfs_program_2(rqstp, transp)
	struct svc_req *rqstp;
	register SVCXPRT *transp;
{
        pThreadParams tp;
        THREAD_TYPE handle;
        pStatInfo stats;

 	if (!CheckNet(transp))
          return;

        stats=StartStatistics("nfsproc");

        tp= AllocThreadQueue();
        tp->rqstp = *rqstp;
        tp->transp = *transp;

	switch (rqstp->rq_proc) {
	case NFSPROC_NULL:
		tp->xdr_argument = xdr_void;
		tp->xdr_result = xdr_void;
		tp->local =  (char *(*)()) nfsproc_null_2;
		break;

	case NFSPROC_GETATTR:
		tp->xdr_argument = xdr_fhandleargs;
		tp->xdr_result = xdr_attrstat;
		tp->local =  (char *(*)()) nfsproc_getattr_2;
		break;

	case NFSPROC_SETATTR:
		tp->xdr_argument = xdr_sattrargs;
		tp->xdr_result = xdr_attrstat;
		tp->local =  (char *(*)()) nfsproc_setattr_2;
		break;

	case NFSPROC_ROOT:
		tp->xdr_argument = xdr_void;
		tp->xdr_result = xdr_void;
		tp->local =  (char *(*)()) nfsproc_root_2;
		break;

	case NFSPROC_LOOKUP:
		tp->xdr_argument = xdr_diropargs;
		tp->xdr_result = xdr_diropres;
		tp->local =  (char *(*)()) nfsproc_lookup_2;
		break;

	case NFSPROC_READLINK:
		tp->xdr_argument = xdr_fhandleargs;
		tp->xdr_result = xdr_readlinkres;
		tp->local =  (char *(*)()) nfsproc_readlink_2;
		break;

	case NFSPROC_READ:
		tp->xdr_argument = xdr_readargs;
		tp->xdr_result = xdr_readres;
		tp->local =  (char *(*)()) nfsproc_read_2;
		break;

	case NFSPROC_WRITECACHE:
		tp->xdr_argument = xdr_void;
		tp->xdr_result = xdr_void;
		tp->local =  (char *(*)()) nfsproc_writecache_2;
		break;

	case NFSPROC_WRITE:
		tp->xdr_argument = xdr_writeargs;
		tp->xdr_result = xdr_attrstat;
		tp->local =  (char *(*)()) nfsproc_write_2;
		break;

	case NFSPROC_CREATE:
		tp->xdr_argument = xdr_createargs;
		tp->xdr_result = xdr_diropres;
		tp->local =  (char *(*)()) nfsproc_create_2;
		break;

	case NFSPROC_REMOVE:
		tp->xdr_argument = xdr_diropargs;
		tp->xdr_result = xdr_nfsstat;
		tp->local =  (char *(*)()) nfsproc_remove_2;
		break;

	case NFSPROC_RENAME:
		tp->xdr_argument = xdr_renameargs;
		tp->xdr_result = xdr_nfsstat;
		tp->local =  (char *(*)()) nfsproc_rename_2;
		break;

	case NFSPROC_LINK:
		tp->xdr_argument = xdr_linkargs;
		tp->xdr_result = xdr_nfsstat;
		tp->local =  (char *(*)()) nfsproc_link_2;
		break;

	case NFSPROC_SYMLINK:
		tp->xdr_argument = xdr_symlinkargs;
		tp->xdr_result = xdr_nfsstat;
		tp->local =  (char *(*)()) nfsproc_symlink_2;
		break;

	case NFSPROC_MKDIR:
		tp->xdr_argument = xdr_createargs;
		tp->xdr_result = xdr_diropres;
		tp->local =  (char *(*)()) nfsproc_mkdir_2;
		break;

	case NFSPROC_RMDIR:
		tp->xdr_argument = xdr_diropargs;
		tp->xdr_result = xdr_nfsstat;
		tp->local =  (char *(*)()) nfsproc_rmdir_2;
		break;

	case NFSPROC_READDIR:
		tp->xdr_argument = xdr_readdirargs;
		tp->xdr_result = xdr_readdirres;
		tp->local =  (char *(*)()) nfsproc_readdir_2;
		break;

	case NFSPROC_STATFS:
		tp->xdr_argument = xdr_fhandleargs;
		tp->xdr_result = xdr_statfsres;
		tp->local =  (char *(*)()) nfsproc_statfs_2;
		break;

	default:
		svcerr_noproc(transp);
                FreeThreadQueue(tp);
                RecordStatistics("nfsproc_noproc",stats);
		return;
	}
	bzero((char *) &(tp->argument), sizeof(tp->argument));

        if (!svc_getargs(transp, tp->xdr_argument, &tp->argument)) {
		svcerr_decode(transp);
                FreeThreadQueue(tp);
                RecordStatistics("nfsproc_badargs",stats);
		return;
	}

        /* the next line is probably NOT PORTABLE */
        tp->udpdata = *(su_data((&tp->transp)));
        tp->pid = getpid();

        if (TestInProgress(tp)) {
          FreeThreadQueue(tp);
          RecordStatistics("nfsproc_inprogress",stats);
        } else {
          AddInProgress(tp);
          AddThreadQueue(tp);
          SIGNAL_SEMAPHORE(SemaphoreRequestWaiting);
          RecordStatistics("nfsproc_signal",stats);
        }

	return;
}

